import unittest
from pymatgen.core import Structure
from importlib import resources
import shutil
import pysipfenn
import pytest
import os

class TestPL(unittest.TestCase):
    """Tests correct loading of the prototype library (used, e.g., for random solid solution generation)."""

    def setUp(self) -> None:
        """Load the prototype library."""
        self.c = pysipfenn.Calculator(autoLoad=False)

    def test_autoload(self):
        """Test that the default prototype library is loaded."""
        self.assertTrue(self.c.prototypeLibrary is not None)
        self.assertTrue(len(self.c.prototypeLibrary) > 0)

    def test_defaultPresent(self):
        """Test that the loaded prototype library was correctly parsed."""
        for prototype in ["FCC", "BCC", "HCP", "Diamond", "DHCP", "Sn_A5"]:
            with self.subTest(msg=prototype):
                self.assertTrue(prototype in self.c.prototypeLibrary)

    def test_correctContentFCC(self):
        """Test that the FCC prototype was correctly parsed."""
        fcc = self.c.prototypeLibrary["FCC"]
        self.assertEqual(fcc["origin"], "https://www.oqmd.org/materials/prototype/A1_Cu")
        self.assertEqual(
            fcc["POSCAR"],
            ('A1_Cu\n'
             '1.0\n'
             '  0.00000   1.80750   1.80750\n'
             '  1.80750   0.00000   1.80750\n'
             '  1.80750   1.80750   0.00000\n'
             'Cu\n'
             '1\n'
             'Direct\n'
             '  0.00000   0.00000   0.00000\n'))
        with self.subTest(msg="Is a pymatgen Structure"):
            self.assertTrue(isinstance(fcc["structure"], Structure))
        with self.subTest(msg="Is valid pymatgen Structure"):
            self.assertTrue(fcc["structure"].is_valid())
        with self.subTest(msg="Has correct formula"):
            self.assertEqual(fcc["structure"].formula, "Cu1")

    def test_customPrototypeLoad(self):
        """Test that a custom prototype can be loaded. Then test that a custom prototype can be appended to the default
        library and stay there."""

        with resources.files('pysipfenn').joinpath('tests/testCaseFiles/prototypeLibrary-custom.yaml') as f:
            self.c.parsePrototypeLibrary(customPath=f, verbose=True, printCustomLibrary=True)

        with self.subTest(msg="Custom prototype present with correct parse"):
            self.assertTrue("NicePhase" in self.c.prototypeLibrary)
            self.assertEqual(self.c.prototypeLibrary["NicePhase"]["origin"], "https://somecustomsource.org")

        with self.subTest(msg="Nice phase is a valid pymatgen Structure"):
            self.assertTrue(isinstance(self.c.prototypeLibrary["NicePhase"]["structure"], Structure))
            self.assertTrue(self.c.prototypeLibrary["NicePhase"]["structure"].is_valid())
            self.assertEqual(self.c.prototypeLibrary["NicePhase"]["structure"].formula, "U1")

        with self.subTest(msg="FCC prototype still present"):
            self.assertTrue("FCC" in self.c.prototypeLibrary)

        with self.subTest(msg="Test that it does not affect the default prototype library"):
            otherC = pysipfenn.Calculator(autoLoad=False)
            self.assertTrue("NicePhase" not in otherC.prototypeLibrary)

        # Create a backup of the default library
        self.c = pysipfenn.Calculator(autoLoad=False)
        backup = self.c.prototypeLibrary.copy()

        with resources.files('pysipfenn').joinpath('tests/testCaseFiles/prototypeLibrary-custom.yaml') as f:
            self.c.appendPrototypeLibrary(customPath=f)

        with self.subTest(msg="Custom prototype present and valid in a different Calculator instance"):
            otherC = pysipfenn.Calculator(autoLoad=False)
            self.assertTrue("NicePhase" in otherC.prototypeLibrary)
            self.assertEqual(otherC.prototypeLibrary["NicePhase"]["origin"], "https://somecustomsource.org")
            self.assertTrue(isinstance(otherC.prototypeLibrary["NicePhase"]["structure"], Structure))
            self.assertTrue(otherC.prototypeLibrary["NicePhase"]["structure"].is_valid())
            self.assertEqual(otherC.prototypeLibrary["NicePhase"]["structure"].formula, "U1")

        with self.subTest(msg="FCC/BCC/HCP prototype still present in a different Calculator instance"):
            self.assertTrue("FCC" in otherC.prototypeLibrary)
            self.assertTrue("BCC" in otherC.prototypeLibrary)
            self.assertTrue("HCP" in otherC.prototypeLibrary)

        with self.subTest(msg="Restore the original prototype library"):
            pysipfenn.overwritePrototypeLibrary(backup)








import unittest
import csv
import os
from pymatgen.core import Structure
from tqdm import tqdm
import numpy as np
from natsort import natsorted
from importlib import resources

from pysipfenn.descriptorDefinitions import KS2022


class TestKS2022(unittest.TestCase):
    '''Tests the correctness of the KS2022 descriptor generation function by comparing the results to the reference data
    for the first 25 structures in the exampleInputFiles directory, stored in the exampleInputFilesDescriptorTable.csv.
    That file that is also used to test the correctness of the Ward2017, which is a superset of the KS2022.
    '''
    def setUp(self):
        '''Reads the reference data from the exampleInputFilesDescriptorTable.csv file and the labels from the first
        row of that file. Then it reads the first 25 structures from the exampleInputFiles directory and generates the
        descriptors for them. The results are stored in the functionOutput list. It defines the emptyLabelsIndx list
        that contains the indices of the labels that are not used in the KS2022 (vs Ward2017) descriptor generation. It
        also persists the test results in the KS2022_TestResult.csv file.
        '''
        with resources.files('pysipfenn'). \
                joinpath('tests/testCaseFiles/exampleInputFilesDescriptorTable.csv').open('r', newline='') as f:
            reader = csv.reader(f)
            referenceDescriptorTable = list(reader)

        self.labels = referenceDescriptorTable[0]
        self.testReferenceData = np.float_(referenceDescriptorTable[1:]).tolist()
        emptyLabels = [
            'mean_WCMagnitude_Shell1', 'mean_WCMagnitude_Shell2', 'mean_WCMagnitude_Shell3',
            'mean_NeighDiff_shell1_SpaceGroupNumber', 'var_NeighDiff_shell1_SpaceGroupNumber',
            'min_NeighDiff_shell1_SpaceGroupNumber', 'max_NeighDiff_shell1_SpaceGroupNumber',
            'range_NeighDiff_shell1_SpaceGroupNumber', 'mean_SpaceGroupNumber', 'maxdiff_SpaceGroupNumber',
            'dev_SpaceGroupNumber', 'max_SpaceGroupNumber', 'min_SpaceGroupNumber', 'most_SpaceGroupNumber',
            'CanFormIonic']
        emptyLabels.reverse()
        self.emptyLabelsIndx = [self.labels.index(l) for l in emptyLabels]

        with resources.files('pysipfenn').joinpath('tests/testCaseFiles/exampleInputFiles/') as exampleInputsDir:
            self.exampleInputFiles = natsorted(os.listdir(exampleInputsDir))
            testStructures = [Structure.from_file(f'{exampleInputsDir}/{eif}') for eif in self.exampleInputFiles]

        self.functionOutput = [KS2022.generate_descriptor(s).tolist() for s in tqdm(testStructures[:25])]
        with resources.files('pysipfenn').joinpath('tests/KS2022_TestResult.csv').open('w+', newline='') as f:
            f.writelines([f'{v}\n' for v in self.functionOutput[0]])

    def test_resutls(self):
        '''Compares the results of the KS2022 descriptor generation function to the reference data on a field-by-field
        basis by calculating the relative difference between the two and requiring it to be less than 1% for all fields
        except 0-valued fields, where the absolute difference is required to be less than 1e-6.
        '''
        for fo, trd, name in zip(self.functionOutput, self.testReferenceData, self.exampleInputFiles):
            for eli in self.emptyLabelsIndx:
                trd.pop(eli)
            for p_fo, p_trd, l in zip(fo, trd, self.labels):
                if p_trd!=0 and p_fo!=0:
                    p_fo_relative = p_fo/p_trd
                    with self.subTest(msg=f'{name:<16} diff in {l}'):
                        self.assertAlmostEqual(p_fo_relative, 1, places=2)
                else:
                    with self.subTest(msg=f'{name:<16} diff in {l}'):
                        self.assertAlmostEqual(p_fo, p_trd, places=6)

    def test_cite(self):
        """Tests citation return."""
        citation = KS2022.cite()
        self.assertIn("Krajewski", citation[0])

class TestKS2022Profiling(unittest.TestCase):
    '''Test the KS2022 descriptor generation by profiling the execution time of the descriptor generation function
    for two example structures (JVASP-10001 and diluteNiAlloy).
    '''
    def test_serial(self):
        '''Test the serial execution of the descriptor generation function 4 times for each of the two examples.'''
        KS2022.profile(test='JVASP-10001', nRuns=4)
        KS2022.profile(test='diluteNiAlloy', nRuns=4)

    def test_parallel(self):
        '''Test the parallel execution of the descriptor generation function 24 times for each of the two examples
        but in parallel with up to 8 workers to speed up the execution.
        '''
        KS2022.profileParallel(test='JVASP-10001', nRuns=24)
        KS2022.profileParallel(test='diluteNiAlloy', nRuns=24)


if __name__ == '__main__':
    unittest.main()



import unittest
import csv
import numpy as np
from importlib import resources
from tqdm.contrib.concurrent import process_map

from pysipfenn.descriptorDefinitions import KS2022_randomSolutions


class TestKSRandomSolution2022(unittest.TestCase):
    def setUp(self):
        '''Load the feature ranges and means from the test file as calculated for the testing/profiling case included
        in the KS2022_randomSolution script run for 50 iterations over BCC Cr12.8 Fe12.8 Co12.8 Ni12.8 Cu12.8 Al35.9
        alloy.
        '''
        testFilePath = 'tests/testCaseFiles/TestFile_DescriptorData_KS2022_randomSolution_valueRangesMeans.csv'
        with resources.files('pysipfenn').joinpath(testFilePath).open('r', newline='') as f:
            reader = csv.reader(f)
            self.descriptorRangeList, self.descriptorMeanList = np.float_(list(reader)).T

        with resources.files('pysipfenn').joinpath('descriptorDefinitions/labels_KS2022.csv').open('r') as f:
            reader = csv.reader(f)
            self.labels = [row[0] for row in reader]

    def test_results(self):
        '''Test the descriptor generation function by comparing the results feature by feature to the reference data
        statistics, testing whether the result is within the observed range of values from the observed mean value
        allowing for additional 2% deviation from the mean value to handle numerical precision in cases where the
        feature converges to near exactly the mean value with near-zero range (e.g. coordination number in BCC in case
        of ideal lattice positions).
        '''
        testValues, meta = KS2022_randomSolutions.profile(
            test='BCC',
            returnDescriptorAndMeta=True,
            plotParameters=True)

        for testValue, descriptorRange, descriptorMean, label in zip(
                testValues,
                self.descriptorRangeList,
                self.descriptorMeanList,
                self.labels):
            with self.subTest(msg=f'{label} in BCC alloy'):
                self.assertGreaterEqual(testValue, (0.95*descriptorMean)-descriptorRange-1e-4)
                self.assertLessEqual(testValue, (1.05*descriptorMean)+descriptorRange+1e-4)

        for field in ['diffHistory', 'propHistory', 'finalAtomsN', 'finalCompositionDistance', 'finalComposition']:
            with self.subTest(msg=f'{field} present in meta'):
                self.assertIn(field, meta)

        with self.subTest(msg="Verify runtime of onlyStructural function to select slices of the descriptor"):
            structuralValues = KS2022_randomSolutions.onlyStructural(testValues)
            self.assertEqual(len(structuralValues), 103)

    def test_errors(self):
        '''Check if correct errors are raised when: (1) the test structure is not implemented.'''
        with self.assertRaises(NotImplementedError):
            KS2022_randomSolutions.profile(test='CrazyStructure')

    def test_cite(self):
        """Tests citation return."""
        citation = KS2022_randomSolutions.cite()
        self.assertIn("Krajewski", citation[0])


class TestKS2022RandomSolutionProfiling(unittest.TestCase):
    '''Test the KS2022 RS descriptor generation by profiling the execution time for example composition for each
    common alloy structure: BCC, FCC, HCP. The profiling is serial inside each task, but done in parallel across
    structures.
    '''
    def test_serialInParallel(self):
        '''Tests profiling a set of structures with parallel task execution.'''
        process_map(KS2022_randomSolutions.profile,
                    ['BCC', 'FCC', 'HCP'],
                    max_workers=3)

    def test_singleInParallel(self):
        '''Tests parallel execution profiling works.'''
        KS2022_randomSolutions.profile(test='BCC', nIterations=2)

if __name__ == '__main__':
    unittest.main()



# Standard Library Imports
import os
import gc
import csv
import json
from time import perf_counter
from typing import List, Union, Dict
from importlib import resources

# Helper Imports
from tqdm import tqdm
from tqdm.contrib.concurrent import process_map
import natsort
from pySmartDL import SmartDL

# Scientific Computing Imports
import numpy as np
from pymatgen.core import Structure, Composition

# Machine Learning Imports
import torch
import onnx2torch
import onnx

# YAML Handling Imports and Configuration
from ruamel.yaml import YAML
from ruamel.yaml.scalarstring import LiteralScalarString

# Descriptor Generators
from pysipfenn.descriptorDefinitions import (
    Ward2017, KS2022, KS2022_dilute, KS2022_randomSolutions
)

# - add new ones here if extending the code

__version__ = '0.15.0'
__authors__ = [["Adam Krajewski", "ak@psu.edu"],
               ["Jonathan Siegel", "jwsiegel@tamu.edu"]]
__name__ = 'pysipfenn'

# *********************************  CALCULATION HIGH-LEVEL ENVIRONMENT  *********************************
class Calculator:
    """pySIPFENN Calculator automatically initializes all functionalities including identification and loading
    of all available models defined statically in the ``models.json`` file. It exposes methods for calculating predefined
    structure-informed descriptors (feature vectors) and predicting properties using models that utilize them.

    Args:
        autoLoad: Automatically load all available ML models based on the ``models.json`` file. This `will` require
            significant memory and time if they are available, so for featurization and other non-model-requiring
            tasks, it is recommended to set this to ``False``. Defaults to ``True``.
        verbose: Print initialization messages and several other non-critical messages during runtime procedures.
            Defaults to True.

    Attributes:
        models: Dictionary with all model information based on the ``models.json`` file in the modelsSIPFENN
            directory. The keys are the network names and the values are dictionaries with the model information.
        loadedModels: Dictionary with all loaded models. The keys are the network names and the values
            are the loaded pytorch models.
        descriptorData: List of all descriptor data created during the last predictions run. The order
            of the list corresponds to the order of atomic structures given to models as input. The order of the
            list of descriptor data for each structure corresponds to the order of networks in the toRun list.
        predictions: List of all predictions created during the last predictions run. The order of the
            list corresponds to the order of atomic structures given to models as input. The order of the list
            of predictions for each structure corresponds to the order of networks in the toRun list.
        inputFiles: List of all input file names used during the last predictions run. The order of the list
            corresponds to the order of atomic structures given to models as input.
    """

    def __init__(self,
                 autoLoad: bool = True,
                 verbose: bool = True):
        """Initializes the pySIPFENN Calculator object."""
        if verbose:
            print('*********  Initializing pySIPFENN Calculator  **********')
            self.verbose = verbose
        # dictionary with all model information
        with resources.files('pysipfenn.modelsSIPFENN').joinpath('models.json').open('r') as f:
            if verbose:
                print(f'Loading model definitions from: {f.name}')
            self.models = json.load(f)
        # networks list
        self.network_list = list(self.models.keys())
        if verbose:
            print(f'Found {len(self.network_list)} network definitions in models.json')
        # network names
        self.network_list_names = [self.models[net]['name'] for net in self.network_list]
        self.network_list_available = []
        self.updateModelAvailability()

        self.loadedModels = {}
        if autoLoad:
            print(f'Loading all available models (autoLoad=True)')
            self.loadModels()
        else:
            print(f'Skipping model loading (autoLoad=False)')

        self.prototypeLibrary = {}
        self.parsePrototypeLibrary(verbose=verbose)

        self.toRun = []
        self.descriptorData = []
        self.predictions = []
        self.metas = {
            'RSS': []
        }
        self.inputFiles = []
        if verbose:
            print(f'*********  pySIPFENN Successfully Initialized  **********')

    def __str__(self):
        """Prints the status of the ``Calculator`` object."""
        printOut = f'pySIPFENN Calculator Object. Version: {__version__}\n'
        printOut += f'Models are located in:\n   {resources.files("pysipfenn.modelsSIPFENN")}\n'
        printOut += f'Auxiliary files (incl. structure prototypes):\n   {resources.files("pysipfenn.misc")}\n{"-" * 80}\n'
        printOut += f'Loaded Networks: {list(self.loadedModels.keys())}\n'
        if len(self.inputFiles) > 0:
            printOut += f'Last files selected as input: {len(self.inputFiles)}\n'
            if len(self.inputFiles) > 4:
                printOut += f'{self.inputFiles[:2]} ... [{len(self.inputFiles) - 4}] ... {self.inputFiles[-2:]}\n'
        if len(self.descriptorData) > 0:
            printOut += f'Last feature calculation run on: {len(self.descriptorData)} structures\n'
        if len(self.toRun) > 0:
            printOut += f'Last Prediction Run Using: {self.toRun}\n'
        if len(self.predictions) > 0:
            printOut += f'Last prediction run on: {len(self.predictions)} structures\n'
            printOut += f'                        {len(self.predictions[0])} predictions/structure\n'
        return printOut

    # *********************************  PROTOTYPE HANDLING  *********************************
    def parsePrototypeLibrary(self,
                              customPath: str = "default",
                              verbose: bool = False,
                              printCustomLibrary: bool = False) -> None:
        """Parses the prototype library YAML file in the ``misc`` directory, interprets them into pymatgen ``Structure``
        objects, and stores them in the ``self.prototypeLibrary`` dict attribute of the ``Calculator`` object. You can use it
        also to temporarily append a custom prototype library (by providing a path) which will live as long as the
        ``Calculator``. For permanent changes, use ``appendPrototypeLibrary()``.

        Args:
            customPath: Path to the prototype library YAML file. Defaults to the magic string ``"default"``, which loads the
                default prototype library included in the package in the ``misc`` directory.
            verbose: If True, it prints the number of prototypes loaded. Defaults to ``False``, but note that ``Calculator``
                class automatically initializes with ``verbose=True``.
            printCustomLibrary: If True, it prints the name and POSCAR of each prototype being added to the prototype
                library. Has no effect if ``customPath`` is ``'default'``. Defaults to ``False``.

        Returns:
            None
        """
        yaml_safeLoader = YAML(typ='safe')

        if customPath == 'default':
            with resources.files('pysipfenn.misc').joinpath('prototypeLibrary.yaml').open('r') as f:
                prototypes = yaml_safeLoader.load(f)
        else:
            with open(customPath, 'r') as f:
                prototypes = yaml_safeLoader.load(f)
                if printCustomLibrary:
                    for prototype in prototypes:
                        print(f'{prototype["name"]}:\n{prototype["POSCAR"]}')
        for prototype in prototypes:
            assert isinstance(prototype['name'], str), 'Prototype name must be a string.'
            assert isinstance(prototype['POSCAR'], str), 'Prototype POSCAR must be a string.'
            assert isinstance(prototype['origin'], str), 'Prototype origin must be a string.'
            struct = Structure.from_str(prototype['POSCAR'], fmt='poscar')
            assert struct.is_valid(), f'Invalid structure for prototype {prototype["name"]}'
            assert struct.is_ordered, f'Unordered structure for prototype {prototype["name"]}. Make sure that the ' \
                                        f'POSCAR file is in the direct format and that no prior randomization has ' \
                                        f'been applied to the structure occupancies.'
            self.prototypeLibrary.update({
                prototype['name']: {
                    'POSCAR': prototype['POSCAR'],
                    'structure': struct,
                    'origin': prototype['origin']
                }
            })
        if verbose:
            print(f'{len(self.prototypeLibrary)} prototype structures present into the prototype library.')

    def appendPrototypeLibrary(self, customPath: str) -> None:
        """Parses a custom prototype library YAML file and permanently appends it into the internal prototypeLibrary
        of the pySIPFENN package. They will be persisted for future use and, by default, they will be loaded
        automatically when instantiating the ``Calculator`` object, similar to your custom models.

        Args:
            customPath: Path to the prototype library YAML file to be appended to the internal ``self.prototypeLibrary``
                of the ``Calculator`` object.

        Returns:
            None
        """

        self.parsePrototypeLibrary(customPath=customPath, printCustomLibrary=True, verbose=True)
        print(f'Now, {len(self.prototypeLibrary)} prototype structures are present into the prototype library. '
              f'Persisting them for future use.')
        overwritePrototypeLibrary(self.prototypeLibrary)

    # *********************************  MODEL HANDLING  *********************************
    def updateModelAvailability(self) -> None:
        """Updates availability of models based on the pysipfenn.modelsSIPFENN directory contents. Works only for
        current ONNX model definitions."""
        with resources.files('pysipfenn.modelsSIPFENN') as p:
            all_files = os.listdir(p)
        detectedNets = []
        for net, netName in zip(self.network_list, self.network_list_names):
            if all_files.__contains__(net + '.onnx'):
                detectedNets.append(net)
                print('✔ ' + netName)
            else:
                print('x ' + netName)
        self.network_list_available = detectedNets

    def downloadModels(self, network: str = 'all') -> None:
        """Downloads ONNX models. By default, all available models are downloaded. If a model is already available
        on disk, it is skipped. If a specific ``network`` is given, only that network is downloaded, possibly overwriting
        the existing one. If the ``network`` name is not recognized, the message will be printed.

        Args:
            network: Name of the network to download. Defaults to ``'all'``.

        """
        with resources.files('pysipfenn.modelsSIPFENN') as modelPath:
            # Fetch all
            if network == 'all':
                print('Fetching all networks!')
                downloadableNets = [net for net in self.network_list if 'URL_ONNX' in self.models[net]]
                for net in self.network_list:
                    if net not in self.network_list_available:
                        if 'URL_ONNX' in self.models[net]:
                            print(f'Fetching: {net}')
                            downloadObject = SmartDL(self.models[net]['URL_ONNX'],
                                                     f'{modelPath}/{net}.onnx',
                                                     threads=16)
                            downloadObject.start()
                            print('\nONNX Network Successfully Fetched.')
                        else:
                            print(f'{net} not detected on disk and ONNX URL has not been provided.')
                    else:
                        print(f'{net} detected on disk. Ready to use.')
                if downloadableNets == self.network_list_available:                
                    print('All downloadable networks are now available!')
                else:
                    print('Problem occurred.')

            # Fetch single
            elif network in self.network_list:
                print(f'Fetching: {network}')
                downloadObject = SmartDL(self.models[network]['URL_ONNX'],
                                         f'{modelPath}/{network}.onnx',
                                         threads=16)
                downloadObject.start()
                print('\nONNX Network Successfully Fetched.')
            # Not recognized
            else:
                print('Network name not recognized')
        self.updateModelAvailability()

    def loadModels(self, network: str = 'all') -> None:
        """Load model/models into memory of the ``Calculator`` class. The models are loaded from the ``modelsSIPFENN`` directory
        inside the package. Its location can be seen by calling ``print()`` on the ``Calculator``. The models are stored in the
        ``self.loadedModels`` attribute as a dictionary with the network string as key and the PyTorch model as value.

         Note:
            This function only works with models that are stored in the ``modelsSIPFENN`` directory inside the package,
            are in ONNX format, and have corresponding entries in ``models.json``. For all others, you will need to use
            ``loadModelCustom()``.

        Args:
            network: Default is ``'all'``, which loads all models detected as available. Alternatively, a specific model
                can be loaded by its corresponding key in models.json. E.g. ``'SIPFENN_Krajewski2020_NN9'`` or
                ``'SIPFENN_Krajewski2022_NN30'``. The key is the same as the network argument in ``downloadModels()``.

        Raises:
            ValueError: If the network name is not recognized or if the model is not available in the ``modelsSIPFENN``
                directory.

        Returns:
            None. It updates the loadedModels attribute of the Calculatorclass.
        """
        with resources.files('pysipfenn.modelsSIPFENN') as modelPath:
            if network == 'all':
                print('Loading models:')
                for net in tqdm(self.network_list_available):
                    self.loadedModels.update({
                        net: onnx2torch.convert(onnx.load(f'{modelPath}/{net}.onnx')).float()
                    })
            elif network in self.network_list_available:
                print('Loading model: ', network)
                self.loadedModels.update({
                    network: onnx2torch.convert(onnx.load(f'{modelPath}/{network}.onnx')).float()
                })
            else:
                raise ValueError(
                    'Network not available. Please check the network name for typos or run downloadModels() '
                    'to download the models. Currently available models are: ', self.network_list_available)

    def loadModelCustom(
            self,
            networkName: str,
            modelName: str,
            descriptor: str,
            modelDirectory: str = '.'
    ) -> None:
        """Load a custom ONNX model from a custom directory specified by the user. The primary use case for this
        function is to load models that are not included in the package and cannot be placed in the package
        directory because of write permissions (e.g. on restrictive HPC systems) or storage allocations.

        Args:
            modelDirectory: Directory where the model is located. Defaults to the current directory.
            networkName: Name of the network. This is the name used to refer to the ONNX network. It has to be
                unique, not contain any spaces, and correspond to the name of the ONNX file (excluding the ``.onnx``
                extension).
            modelName: Name of the model. This is the name that will be displayed in the model selection menu. It
                can be any string desired.
            descriptor: Descriptor/feature vector used by the model. pySIPFENN currently supports the following
                descriptors: ``'KS2022'``, and ``'Ward2017'``.
        """

        self.loadedModels.update({
            networkName: onnx2torch.convert(onnx.load(f'{modelDirectory}/{networkName}.onnx')).float()
        })
        self.models.update({
            networkName: {
                'name': modelName,
                'descriptor': descriptor
            }})
        self.network_list.append(networkName)
        self.network_list_names.append(modelName)
        self.network_list_available.append(networkName)
        print(f'Loaded model {modelName} ({networkName}) from {modelDirectory}')

    def findCompatibleModels(self, descriptor: str) -> List[str]:
        """Finds all models compatible with a given descriptor based on the descriptor definitions loaded from the
        ``models.json`` file.

        Args:
            descriptor: Descriptor to use. Must be one of the available descriptors. See ``pysipfenn.descriptorDefinitions``
                to see available modules or add yours. Available default descriptors are: ``'Ward2017'``, ``'KS2022'``.

        Returns:
            List of strings corresponding to compatible models.
        """

        compatibleList = []
        for net in self.models:
            if descriptor in self.models[net]['descriptor']:
                compatibleList.append(net)
        return compatibleList

    # *******************************  DESCRIPTOR HANDLING (MID-LEVEL API) *******************************
    def calculate_Ward2017(
            self,
            structList: List[Structure],
            mode: str = 'serial',
            max_workers: int = 4
    ) -> list:
        """Calculates ``Ward2017`` descriptors for a list of structures. The calculation can be done in serial or parallel
        mode. In parallel mode, the number of workers can be specified. The results are stored in the ``self.descriptorData``
        attribute. The function returns the list of descriptors as well.

        Args:
            structList: List of structures to calculate descriptors for. The structures must be
                initialized with the pymatgen ``Structure`` class.
            mode: Mode of calculation. Defaults to 'serial'. Options are ``'serial'`` and ``'parallel'``.
            max_workers: Number of workers to use in parallel mode. Defaults to ``4``. If ``None``, the number of workers
                will be set to the number of available CPU cores. If set to ``0``, 1 worker will be used.

        Returns:
            List of ``Ward2017`` descriptor (feature vector) for each structure.

        """
        if mode == 'serial':
            descList = [Ward2017.generate_descriptor(s) for s in tqdm(structList)]
            if self.verbose: print('Done!')
            self.descriptorData = descList
            return descList
        elif mode == 'parallel':
            descList = process_map(Ward2017.generate_descriptor, structList, max_workers=max_workers)
            if self.verbose: print('Done!')
            self.descriptorData = descList
            return descList

    def calculate_KS2022(
            self,
            structList: List[Structure],
            mode: str = 'serial',
            max_workers: int = 8
    ) -> list:
        """Calculates ``KS2022`` descriptors for a list of structures. The calculation can be done in serial or parallel
        mode. In parallel mode, the number of workers can be specified. The results are stored in the descriptorData
        attribute. The function returns the list of descriptors as well.

        Args:
            structList: List of structures to calculate descriptors for. The structures must be
                initialized with the pymatgen ``Structure`` class.
            mode: Mode of calculation. Defaults to 'serial'. Options are ``'serial'`` and ``'parallel'``.
            max_workers: Number of workers to use in parallel mode. Defaults to ``8``. If ``None``, the number of workers
                will be set to the number of available CPU cores. If set to ``0``, 1 worker will be used.

        Returns:
            List of ``KS2022`` descriptor (feature vector) for each structure.

        """
        if mode == 'serial':
            descList = [KS2022.generate_descriptor(s) for s in tqdm(structList)]
            if self.verbose: print('Done!')
            self.descriptorData = descList
            return descList
        elif mode == 'parallel':
            descList = process_map(KS2022.generate_descriptor, structList, max_workers=max_workers)
            if self.verbose: print('Done!')
            self.descriptorData = descList
            return descList

    def calculate_KS2022_dilute(
            self,
            structList: List[Structure],
            baseStruct: Union[str, List[Structure]] = 'pure',
            mode: str = 'serial',
            max_workers: int = 8
    ) -> List[np.ndarray]:
        """Calculates ``KS2022`` descriptors for a list of dilute structures (either based on pure elements and on custom
        base structures, e.g. TCP endmember configurations) that contain a single alloying atom. Speed increases are
        substantial compared to the ``KS2022`` descriptor, which is more general and can be used on any structure. The
        calculation can be done in serial or parallel mode. In parallel mode, the number of workers can be specified.
        The results are stored in the ``self.descriptorData`` attribute. The function returns the list of descriptors as well.

        Args:
            structList: List of structures to calculate descriptors for. The structures must be
                dilute structures (either based on pure elements and on custom base structures, e.g. TCP endmember
                configurations) that contain a single alloying atom. The structures must be initialized with the
                pymatgen ``Structure`` class.
            baseStruct: Non-diluted references for the dilute structures. Defaults to ``'pure'``, which assumes that the structures
                are based on pure elements and generates references automatically. Alternatively, a list of structures
                can be provided, which can be either pure elements or custom base structures (e.g. TCP endmember
                configurations).
            mode: Mode of calculation. Defaults to ``'serial'``. Options are ``'serial'`` and ``'parallel'``.
            max_workers: Number of workers to use in parallel mode. Defaults to ``8``. If ``None``, the number of workers
                will be set to the number of available CPU cores. If set to ``0``, 1 worker will be used.

        Returns:
            List of ``KS2022`` descriptor (feature vector) ``np.ndarray`` for each structure.
        """

        if baseStruct == 'pure' or isinstance(baseStruct, Structure):
            if mode == 'serial':
                descList = [KS2022_dilute.generate_descriptor(s, baseStruct=baseStruct) for s in tqdm(structList)]
                if self.verbose:
                    print('Done!')
                self.descriptorData = descList
                return descList
            elif mode == 'parallel':
                pairedInput = list(zip(structList, [baseStruct] * len(structList)))
                descList = process_map(wrapper_KS2022_dilute_generate_descriptor,
                                       pairedInput,
                                       max_workers=max_workers)
                if self.verbose:
                    print('Done!')
                self.descriptorData = descList
                return descList

        elif isinstance(baseStruct, List) and len(baseStruct) == len(structList):
            if mode == 'serial':
                descList = [KS2022_dilute.generate_descriptor(s, bs) for s, bs in tqdm(zip(structList, baseStruct))]
                if self.verbose:
                    print('Done!')
                self.descriptorData = descList
                return descList
            elif mode == 'parallel':
                pairedInput = list(zip(structList, baseStruct))
                descList = process_map(wrapper_KS2022_dilute_generate_descriptor,
                                       pairedInput, max_workers=max_workers)
                if self.verbose:
                    print('Done!')
                self.descriptorData = descList
                return descList
            else:
                raise ValueError('`baseStruct` must be (1) `pure`, (2) `Structure` or a list of them.')

    def calculate_KS2022_randomSolutions(
            self,
            baseStructList: Union[str, Structure, List[str], List[Structure], List[Union[Composition, str]]],
            compList: Union[str, List[str], Composition, List[Composition], List[Union[Composition, str]]],
            minimumSitesPerExpansion: int = 50,
            featureConvergenceCriterion: float = 0.005,
            compositionConvergenceCriterion: float = 0.01,
            minimumElementOccurrences: int = 10,
            plotParameters: bool = False,
            printProgress: bool = False,
            mode: str = 'serial',
            max_workers: int = 8
    ) -> List[np.ndarray]:
        """Calculates ``KS2022`` descriptors corresponding to random solid solutions occupying base structure / lattice
        sites for a list of compositions through method described in ``descriptorDefinitions.KS2022_randomSolutions``
        submodule. The results are stored in the descriptorData attribute. The function returns the list of descriptors
        in numpy format as well.

        Args:
            baseStructList: The base structure to generate a random solid solution (RSS). It does _not_ need to be a simple
                Bravis lattice, such as BCC lattice, but can be any ``Structure`` object or a list of them, if you need to
                define them on per-case basis. In addition to `Structure` objects, you can use "magic" strings
                corresponding to one of the structures in the library you can find under ``pysipfenn.misc`` directory or
                loaded under ``self.prototypeLibrary`` attribute. The magic strings include, but are not limited to:
                ``'BCC'``, ``'FCC'``, ``'HCP'``, ``'DHCP'``, ``'Diamond'``, and so on. You can invoke them by their name, e.g. ``BCC``, or
                by passing ``self.prototypeLibrary['BCC']['structure']`` directly. If you pass a list to ``baseStruct``,
                you are allowed to mix-and-match ``Structure`` objects and magic strings.
            compList: The composition to populate the supercell with until KS2022 descriptor converges. You can use
                pymatgen's ``Composition`` objects or strings of valid chemical formulas (symbol - atomic fraction pairs),
                like ``'Fe0.5Ni0.3Cr0.2'``, ``'Fe50 Ni30 Cr20'``, or ``'Fe5 Ni3 Cr2'``. You can either pass a single entity, in
                which case it will be used for all structures (use to run the same composition for different base
                structures), or a list of entities, in which case pairs will be used in the order of the list. If you
                pass a list to ``compList``, you are allowed to mix-and-match ``Composition`` objects and composition
                strings.
            minimumSitesPerExpansion: The minimum number of sites that the base structure will be expanded to (doubling
                dimension-by-dimension) before it is used as expansion step/batch in each iteration of adding local
                chemical environment information to the global ensemble.
                The optimal value will depend on the number of species and their relative fractions in the composition.
                Generally, low values (<20ish) will result in a slower convergence, as some extreme local chemical
                environments will have strong influence on the global ensemble, and too high values (>150ish) will
                result in a needlessly slow computation for not-complex compositions, as at least two iterations will
                be processed. The default value is ``50`` and works well for simple cases.
            featureConvergenceCriterion: The maximum difference between any feature belonging to the current iteration
                (statistics based on the global ensemble of local chemical environments) and the previous iteration
                (before last expansion) expressed as a fraction of the maximum value of each feature found in the OQMD
                database at the time of SIPFENN creation (see ``KS2022_randomSolutions.maxFeaturesInOQMD`` array).
                The default value is ``0.01``, corresponding to 1% of the maximum value.
            compositionConvergenceCriterion: The maximum average difference between any element fraction belonging to
                the current composition (net of all expansions) and the target composition (``comp``). The default value
                is ``0.01``, corresponding to 1% deviation, which interpretation will depend on the number of elements
                in the composition.
            minimumElementOccurrences: The minimum number of times all elements must occur in the composition before it
                is considered converged. This setting prevents the algorithm from converging before very dilute elements
                like C in low-carbon steel, have had a chance to occur. The default value is ``10``.
            plotParameters: If True, the convergence history will be plotted using plotly. The default value is ``False``,
                but tracking them is recommended and will be accessible in the `metas` attribute of the Calculator under
                the key ``'RSS'``.
            printProgress: If True, the progress will be printed to the console. The default value is False.
            mode: Mode of calculation. Options are ``serial`` (default) and ``parallel``.
            max_workers: Number of workers to use in parallel mode. Defaults to ``8``.

        Returns:
            A list of ``numpy.ndarray``s containing the ``KS2022`` descriptor, just like the ordinary ``KS2022``. **Please note
            the stochastic nature of this algorithm**. The result will likely vary slightly between runs and parameters,
            so if convergence is critical, verify it with a test matrix of ``minimumSitesPerExpansion``,
            ``featureConvergenceCriterion``, and ``compositionConvergenceCriterion`` values.
        """
        # LIST-LIST: Assert that if both baseStruct and compList are lists, they have the same length
        if isinstance(baseStructList, list) and isinstance(compList, list):
            assert len(baseStructList) == len(compList), \
                'baseStruct and compList must have the same length if both are lists. If you want to use the same ' \
                'entity for all calculations, do not wrap it.'

        # STRING / STRUCT handling and extension
        if isinstance(baseStructList, str) or isinstance(baseStructList, Structure):
            baseStructList = [baseStructList]
            if isinstance(compList, list) and len(compList) > 1:
                baseStructList = baseStructList * len(compList)
        else:
            assert isinstance(baseStructList, list), 'baseStruct must be a list if it is not a string or Structure.'

        if isinstance(compList, str) or isinstance(compList, Composition):
            compList = [compList]
            if isinstance(baseStructList, list) and len(baseStructList) > 1:
                compList = compList * len(baseStructList)
        else:
            assert isinstance(compList, list), 'compList must be a list if it is not a string or Composition.'

        # LISTS of STRING / STRUCT
        for i in range(len(baseStructList)):
            assert isinstance(baseStructList[i], (str, Structure)), \
                'baseStruct must be a list of strings or Structure objects.'
            if isinstance(baseStructList[i], str):
                baseStructList[i] = string2prototype(self, baseStructList[i])

        for i in range(len(compList)):
            assert isinstance(compList[i], (str, Composition)), \
                'compList must be a list of strings or Composition objects.'
            if isinstance(compList[i], str):
                c = Composition(compList[i])
                assert c.valid, f'Unrecognized composition string: {compList}. Please provide a valid composition ' \
                                f'string, e.g. "Fe0.5Ni0.3Cr0.2", "Fe50 Ni30 Cr20", or "Fe5 Ni3 Cr2".'
                compList[i] = c

        assert len(baseStructList) == len(compList), 'baseStruct and compList must have the same length at this point.'
        pairedInputAndSettings, descList, metaList = [], [], []

        for i in range(len(baseStructList)):
            pairedInputAndSettings.append(
                (baseStructList[i],
                 compList[i],
                 minimumSitesPerExpansion,
                 featureConvergenceCriterion,
                 compositionConvergenceCriterion,
                 minimumElementOccurrences,
                 plotParameters,
                 printProgress,
                 True))

        if mode == 'serial':
            for base, comp, *settings in tqdm(pairedInputAndSettings):
                desc, meta = KS2022_randomSolutions.generate_descriptor(base, comp, *settings)
                descList.append(desc)
                metaList.append(meta)

        elif mode == 'parallel':
            print(pairedInputAndSettings)
            descList, metaList = zip(*process_map(
                wrapper_KS2022_randomSolutions_generate_descriptor,
                pairedInputAndSettings,
                max_workers=max_workers
            ))
        else:
            raise ValueError('Incorrect calculation mode selected. Must be either `serial` or `parallel`.')

        if self.verbose:
            print('Done!')
        self.descriptorData = descList
        self.metas['RSS'] = metaList
        return descList

    # *******************************  PREDICTION RUNNERS (MID-LEVEL API) *******************************
    def makePredictions(
            self,
            models: Dict[str, torch.nn.Module],
            toRun: List[str],
            dataInList: List[Union[List[float], np.array]]
    ) -> List[list]:
        """Makes predictions using PyTorch networks listed in toRun and provided in models dictionary. Shared among all
        "predict" functions.

        Args:
            models: Dictionary of models to use. Keys are network names and values are PyTorch models loaded from ONNX
                with ``loadModels()`` / ``loadModelCustom()`` or manually (fairly simple!).
            toRun: List of networks to run. It must be a subset of ``models.keys()``.
            dataInList: List of data to make predictions for. Each element of the list should be a descriptor accepted
                by all networks in toRun. Can be a list of lists of floats or a list of numpy ``nd.array``s.

        Returns:
            List of predictions. Each element of the list is a list of predictions for all run networks. The order of the
            predictions is the same as the order of the networks in ``toRun``.
        """
        dataOuts = []
        if self.verbose:
            print('Making predictions...')
        # Run for each network
        dataIn = torch.from_numpy(np.array(dataInList)).float()
        assert set(toRun).issubset(set(models.keys())), 'Some networks to run are not available in the models.'
        for net in toRun:
            t0 = perf_counter()
            model = models[net]
            model.eval()
            if 'OnnxDropoutDynamic()' in {str(module) for module in list(model._modules.values())}:
                tempOut = model(dataIn, None)
            else:
                tempOut = model(dataIn)
            t1 = perf_counter()
            dataOuts.append(tempOut.cpu().detach().numpy())
            if self.verbose:
                print(f'Prediction rate: {round(len(tempOut) / (t1 - t0), 1)} pred/s')
                print(f'Obtained {len(tempOut)} predictions from:  {net}')

        # Transpose and round the predictions
        dataOuts = np.array(dataOuts).T.tolist()[0]
        self.predictions = dataOuts
        return dataOuts

    # *******************************  TOP-LEVEL API  *******************************
    def runModels(
            self,
            descriptor: str,
            structList: List[Structure],
            mode: str = 'serial',
            max_workers: int = 4
    ) -> List[List[float]]:
        """Runs all loaded models on a list of Structures using specified descriptor. Supports serial and parallel
        computation modes. If parallel is selected, max_workers determines number of processes handling the
        featurization of structures (90-99+% of computational intensity) and models are then run in series.

        Args:
            descriptor: Descriptor to use. Must be one of the available descriptors. See ``pysipfenn.descriptorDefinitions``
                to see available modules or add yours. Available default descriptors are: ``'Ward2017'``, ``'KS2022'``.
            structList: List of pymatgen Structure objects to run the models on.
            mode: Computation mode. ``'serial'`` or ``'parallel'``. Default is ``'serial'``. Parallel mode is not recommended for
                small datasets.
            max_workers: Number of workers to use in parallel mode. Default is ``4``. Ignored in serial mode. If set to
                ``None``, will use all available cores. If set to ``0``, will use ``1`` core.

        Returns:
            List of predictions. Each element of the list is a list of predictions for all ran networks. The
            order of the predictions is the same as the order of the input structures. The order of the networks is
            the same as the order of the networks in ``self.network_list_available``. If a network is not available, it
            will not be included in the list. If a network is not compatible with the selected descriptor, it will
            not be included in the list.
        """

        self.toRun = list(set(self.findCompatibleModels(descriptor)).intersection(set(self.network_list_available)))
        self.toRun = natsort.natsorted(self.toRun)
        if len(self.toRun) == 0:
            print('The list of models to run is empty. This may be caused by selecting a descriptor not '
                  'defined/available, or if the selected descriptor does not correspond to any available network. '
                  'Check spelling and invoke the downloadModels() function if you are using base models.')
            raise AssertionError
        else:
            print(f'\nModels that will be run: {self.toRun}')

        print('Calculating descriptors...')
        if descriptor == 'Ward2017':
            self.descriptorData = self.calculate_Ward2017(
                structList=structList,
                mode=mode,
                max_workers=max_workers
            )
        elif descriptor == 'KS2022':
            self.descriptorData = self.calculate_KS2022(
                structList=structList,
                mode=mode,
                max_workers=max_workers
            )
        else:
            print('Descriptor handing not implemented. Check spelling.')
            raise AssertionError

        self.makePredictions(
            models=self.loadedModels,
            toRun=self.toRun,
            dataInList=self.descriptorData
        )

        return self.predictions

    def runModels_dilute(
            self,
            descriptor: str,
            structList: List[Structure],
            baseStruct: Union[str, List[Structure]] = 'pure',
            mode: str = 'serial',
            max_workers: int = 4
    ) -> List[List[float]]:
        """Runs all loaded models on a list of Structures using specified descriptor. A critical difference
        from runModels() is that this function will call dilute-specific featurizer, e.g. ``KS2022_dilute`` when ``'KS2022'`` is
        provided as input, which can only be used on dilute structures (both based on pure elements and on custom base
        structures, e.g. TCP endmember configurations) that contain a single alloying atom. Speed increases are
        substantial compared to the KS2022 descriptor, which is more general and can be used on any structure.
        Supports serial and parallel modes in the same way as ``runModels()``.

        Args:
            descriptor: Descriptor to use for predictions. Must be one of the descriptors which support the dilute
                structures (i.e. `*_dilute`). See ``pysipfenn.descriptorDefinitions`` to see available modules or add yours
                here. Available default dilute descriptors are now: ``'KS2022'``. The ``'KS2022'`` can also be called from
                ``runModels()`` function, but is not recommended for dilute alloys, as it negates the speed increase of the
                dilute structure featurizer.
            structList: List of pymatgen ``Structure`` objects to run the models on. Must be dilute structures as described
                above.
            baseStruct: Non-diluted references for the dilute structures. Defaults to 'pure', which assumes that the
                structures are based on pure elements and generates references automatically. Alternatively, a list of
                structures can be provided, which can be either pure elements or custom base structures (e.g. TCP
                endmember configurations).
            mode: Computation mode. ``'serial'`` or ``'parallel'``. Default is ``'serial'``. Parallel mode is not recommended for
                small datasets.
            max_workers: Number of workers to use in parallel mode. Default is ``4``. Ignored in serial mode. If set to
                ``None``, will use all available cores. If set to ``0``, will use ``1`` core.

        Returns:
            List of predictions. Each element of the list is a list of predictions for all run networks. The
            order of the predictions is the same as the order of the input structures. The order of the networks
            is the same as the order of the networks in ``self.network_list_available``. If a network is not available,
            it will not be included in the list. If a network is not compatible with the selected descriptor, it
            will not be included in the list.
        """

        self.toRun = list(set(self.findCompatibleModels(descriptor)).intersection(set(self.network_list_available)))
        if len(self.toRun) == 0:
            print('The list of models to run is empty. This may be caused by selecting a descriptor not '
                  'defined/available, or if the selected descriptor does not correspond to any available network. '
                  'Check spelling and invoke the downloadModels() function if you are using base models.')
            raise AssertionError
        else:
            print(f'Running {self.toRun} models')

        print('Calculating descriptors...')
        if descriptor == 'KS2022':
            self.descriptorData = self.calculate_KS2022_dilute(
                structList=structList,
                baseStruct=baseStruct,
                mode=mode,
                max_workers=max_workers
            )
        else:
            print('Descriptor handing not implemented. Check spelling.')
            raise AssertionError

        self.makePredictions(
            models=self.loadedModels,
            toRun=self.toRun,
            dataInList=self.descriptorData
        )

        return self.predictions

    def runModels_randomSolutions(
            self,
            descriptor: str,
            baseStructList: Union[str, Structure, List[str], List[Structure], List[Union[Composition, str]]],
            compList: Union[str, List[str], Composition, List[Composition], List[Union[Composition, str]]],
            minimumSitesPerExpansion: int = 50,
            featureConvergenceCriterion: float = 0.005,
            compositionConvergenceCriterion: float = 0.01,
            minimumElementOccurrences: int = 10,
            plotParameters: bool = False,
            printProgress: bool = False,
            mode: str = 'serial',
            max_workers: int = 8,
        ) -> List[List[float]]:
        """A top-level convenience wrapper for the ``calculate_KS2022_randomSolutions`` function. It passes all the
        arguments to that function directly (except for ``descriptor`` and uses its result to run all applicable models.
        The result is a list of predictions for all run networks.

        Args:
            descriptor: Descriptor to use for predictions. Must be one of the descriptors which support the random
            solid solution structures (i.e. `*_randomSolutions`). See ``pysipfenn.descriptorDefinitions`` to see
            available modules or add yours here. As of v0.15.0, the only available descriptor is
            ``'KS2022'`` through its ``KS2022_randomSolutions`` submodule.
            baseStructList: See ``calculate_KS2022_randomSolutions`` for details. You can mix-and-match ``Structure``
                objects and magic strings, either individually (to use the same entity for all calculations) or in a
                list.
            compList: See ``calculate_KS2022_randomSolutions`` for details. You can mix-and-match ``Composition``
                objects and composition strings, either individually (to use the same entity for all calculations)
                or in a list.
            minimumSitesPerExpansion: See ``calculate_KS2022_randomSolutions``.
            featureConvergenceCriterion: See ``calculate_KS2022_randomSolutions``.
            compositionConvergenceCriterion: See ``calculate_KS2022_randomSolutions``.
            minimumElementOccurrences: See ``calculate_KS2022_randomSolutions``.
            plotParameters: See ``calculate_KS2022_randomSolutions``.
            printProgress: See ``calculate_KS2022_randomSolutions``.
            mode: Computation mode. ``'serial'`` or ``'parallel'``. Default is ``'serial'``. Parallel mode is not
                recommended for small datasets.

        Returns:
            List of predictions. They will correspond to the order of the networks in ``self.toRun`` established by the
            ``findCompatibleModels()`` function. If a network is not available, it will not be included in the list.
        """

        self.toRun = list(set(self.findCompatibleModels(descriptor)).intersection(set(self.network_list_available)))
        if len(self.toRun) == 0:
            print('The list of models to run is empty. This may be caused by selecting a descriptor not '
                  'defined/available, or if the selected descriptor does not correspond to any available network. '
                  'Check spelling and invoke the downloadModels() function if you are using base models.')
            raise AssertionError
        else:
            print(f'Running {self.toRun} models')

        print('Calculating descriptors...')
        if descriptor == 'KS2022':
            self.descriptorData = self.calculate_KS2022_randomSolutions(
                baseStructList=baseStructList,
                compList=compList,
                minimumSitesPerExpansion=minimumSitesPerExpansion,
                featureConvergenceCriterion=featureConvergenceCriterion,
                compositionConvergenceCriterion=compositionConvergenceCriterion,
                minimumElementOccurrences=minimumElementOccurrences,
                plotParameters=plotParameters,
                printProgress=printProgress,
                mode=mode,
                max_workers=max_workers
            )
        else:
            print('Descriptor handing not implemented. Check spelling.')
            raise AssertionError

        self.makePredictions(
            models=self.loadedModels,
            toRun=self.toRun,
            dataInList=self.descriptorData)

        return self.predictions

    def runFromDirectory(
            self,
            directory: str,
            descriptor: str,
            mode: str = 'serial',
            max_workers: int = 4
    ) -> List[list]:
        """Runs all loaded models on a list of Structures it automatically imports from a specified directory. The
        directory must contain only atomic structures in formats such as ``'poscar'``, ``'cif'``, ``'json'``, ``'mcsqs'``, etc.,
        or a mix of these. The structures are automatically sorted using natsort library, so the order of the
        structures in the directory, as defined by the operating system, is not important. Natural sorting,
        for example, will sort the structures in the following order: ``'1-Fe'``, ``'2-Al'``, ``'10-xx'``, ``'11-xx'``, ``'20-xx'``,
        ``'21-xx'``, ``'11111-xx'``, etc. This is useful when the structures are named using a numbering system. The order of
        the predictions is the same as the order of the input structures. The order of the networks in a prediction
        is the same as the order of the networks in ``self.network_list_available``. If a network is not available,
        it will not be included in the list.

        Args:
            directory: Directory containing the structures to run the models on. The directory must contain only atomic
                structures in formats such as ``'poscar'``, ``'cif'``, ``'json'``, ``'mcsqs'``, etc., or a mix of these. The structures
                are automatically sorted as described above.
            descriptor: Descriptor to use. Must be one of the available descriptors. See ``pysipgenn.descriptorDefinitions``
                for a list of available descriptors.
            mode: Computation mode. ``'serial'`` or ``'parallel'``. Default is ``'serial'``. Parallel mode is not recommended for
                small datasets.
            max_workers: Number of workers to use in parallel mode. Default is ``4``. Ignored in serial mode. If set to
                ``None``, will use all available cores. If set to ``0``, will use 1 core.

        Returns:
            List of predictions. Each element of the list is a list of predictions for all run networks. The order of
            the predictions is the same as the order of the input structures. The order of the networks is the same as
            the order of the networks in ``self.network_list_available``. If a network is not available, it will not be
            included in the list.
        """

        print('Importing structures...')
        self.inputFiles = os.listdir(directory)
        self.inputFiles = natsort.natsorted(self.inputFiles)
        structList = [Structure.from_file(f'{directory}/{eif}') for eif in tqdm(self.inputFiles)]
        self.runModels(descriptor=descriptor, structList=structList, mode=mode, max_workers=max_workers)
        print('Done!')

        return self.predictions

    def runFromDirectory_dilute(self,
                                directory: str,
                                descriptor: str,
                                baseStruct: str = 'pure',
                                mode: str = 'serial',
                                max_workers: int = 8) -> None:
        """Runs all loaded models on a list of dilute Structures it automatically imports from a specified directory.
        The directory must contain only atomic structures in formats such as ``'poscar'``, ``'cif'``, ``'json'``, ``'mcsqs'``, etc.,
        or a mix of these. The structures are automatically sorted using natsort library, so the order of the
        structures in the directory, as defined by the operating system, is not important. Natural sorting,
        for example, will sort the structures in the following order: ``'1-Fe'``, ``'2-Al'``, ``'10-xx'``, ``'11-xx'``, ``'20-xx'``,
        ``'21-xx'``, ``'11111-xx'``, etc. This is useful when the structures are named using a numbering system. The order of
        the predictions is the same as the order of the input structures. The order of the networks in a prediction
        is the same as the order of the networks in self.network_list_available. If a network is not available,
        it will not be included in the list.

        Args:
            directory: Directory containing the structures to run the models on. The directory must contain only atomic
                structures in formats such as ``'poscar'``, ``'cif'``, ``'json'``, ``'mcsqs'``, etc., or a mix of these. The structures
                are automatically sorted as described above. The structures must be dilute structures, i.e. they must
                contain only one alloying element.
            descriptor: Descriptor to use. Must be one of the available descriptors. See ``pysipfenn.descriptorDefinitions``
                for a list of available descriptors.
            baseStruct: Non-diluted references for the dilute structures. Defaults to ``'pure'``, which assumes that the
                structures are based on pure elements and generates references automatically. Alternatively, a list of
                structures can be provided, which can be either pure elements or custom base structures (e.g. TCP
                endmember configurations).
            mode: Computation mode. ``'serial'`` or ``'parallel'``. Default is ``'serial'``. Parallel mode is not recommended for
                small datasets.
            max_workers: Number of workers to use in parallel mode. Default is ``8``. Ignored in serial mode. If set to
                ``None``, will use all available cores. If set to ``0``, will use 1 core.

        Returns:
            List of predictions. Each element of the list is a list of predictions for all run networks. The order of
            the predictions is the same as the order of the input structures. The order of the networks is the same as
            the order of the networks in ``self.network_list_available``. If a network is not available, it will not be
            included in the list.

        """
        print('Importing structures...')
        self.inputFiles = os.listdir(directory)
        self.inputFiles = natsort.natsorted(self.inputFiles)
        structList = [Structure.from_file(f'{directory}/{eif}') for eif in tqdm(self.inputFiles)]
        self.runModels_dilute(descriptor=descriptor,
                              structList=structList,
                              baseStruct=baseStruct,
                              mode=mode,
                              max_workers=max_workers)
        print('Done!')


    # *******************************  POST-PROCESSING  *******************************
    def get_resultDicts(self) -> List[dict]:
        """Returns a list of dictionaries with the predictions for each network. The keys of the dictionaries are the
        names of the networks. The order of the dictionaries is the same as the order of the input structures passed
        through ``runModels()`` functions.

        Returns:
            List of dictionaries with the predictions.
        """
        return [dict(zip(self.toRun, pred)) for pred in self.predictions]

    def get_resultDictsWithNames(self) -> List[dict]:
        """Returns a list of dictionaries with the predictions for each network. The keys of the dictionaries are the
        names of the networks and the names of the input structures. The order of the dictionaries is the same as the
        order of the input structures passed through ``runModels()`` functions. Note that this function requires
        ``self.inputFiles`` to be set, which is done automatically when using ``runFromDirectory()`` or
        ``runFromDirectory_dilute()`` but not when using ``runModels()`` or ``runModels_dilute()``, as the input structures are
        passed directly to the function and names have to be provided separately by assigning them to ``self.inputFiles``.

        Returns:
            List of dictionaries with the predictions.
        """
        assert self.inputFiles is not []
        assert len(self.inputFiles) == len(self.predictions)
        return [
            dict(zip(['name'] + self.toRun, [name] + pred))
            for name, pred in
            zip(self.inputFiles, self.predictions)]


    def writeResultsToCSV(self, file: str) -> None:
        """Writes the results to a CSV file. The first column is the name of the structure. If the ``self.inputFiles``
        attribute is populated automatically by ``runFromDirectory()`` or set manually, the names of the structures will
        be used. Otherwise, the names will be ``'1'``, ``'2'``, ``'3'``, etc. The remaining columns are the predictions for each
        network. The order of the columns is the same as the order of the networks in ``self.network_list_available``.

        Args:
            file: Name of the file to write the results to. If the file already exists, it will be overwritten. If the
                file does not exist, it will be created. The file must have a ``'.csv'`` extension to be recognized
                correctly.
        """

        assert self.toRun is not []
        with open(file, 'w+', encoding="utf-8") as f:
            f.write('Name,' + ','.join(self.toRun) + '\n')
            if len(self.inputFiles) == len(self.predictions):
                for name, pred in zip(self.inputFiles, self.predictions):
                    assert len(pred) == len(self.toRun)
                    f.write(f'{name},{",".join(str(v) for v in pred)}\n')
            else:
                i = 1
                for pred in self.predictions:
                    f.write(f'{i},{",".join(str(v) for v in pred)}\n')
                    i += 1

    def writeDescriptorsToCSV(self, descriptor: str, file: str = 'descriptorData.csv') -> None:
        """Writes the descriptor data to a CSV file. The first column is the name of the structure. If the
        ``self.inputFiles`` attribute is populated automatically by runFromDirectory() or set manually, the names of the
        structures will be used. Otherwise, the names will be ``'1'``, ``'2'``, ``'3'``, etc. The remaining columns are the
        descriptor values. The order of the columns is the same as the order of the labels in the descriptor
        definition file.

        Args:
            descriptor: Descriptor to use. Must be one of the available descriptors. See ``pysipgenn.descriptorDefinitions``
                for a list of available descriptors, such as ``'KS2022'`` and ``'Ward2017'``. It provides the labels for the
                descriptor values.
            file: Name of the file to write the results to. If the file already exists, it will be overwritten. If the
                file does not exist, it will be created. The file must have a ``'.csv'`` extension to be recognized
                correctly.
        """

        # Load descriptor labels
        with resources.files('pysipfenn').joinpath(f'descriptorDefinitions/labels_{descriptor}.csv') as labelsCSV:
            with open(labelsCSV, 'r') as f:
                reader = csv.reader(f)
                labels = [v[0] for v in list(reader)]

        # Write descriptor data
        with open(file, 'w+', encoding="utf-8") as f:
            f.write(f'Name,{",".join(labels)}\n')
            if len(self.inputFiles) == len(self.descriptorData):
                for name, dd in zip(self.inputFiles, self.descriptorData):
                    assert len(dd) == len(labels)
                    f.write(f'{name},{",".join(str(v) for v in dd.tolist())}\n')
            else:
                i = 1
                for dd in self.descriptorData:
                    f.write(f'{i},{",".join(str(v) for v in dd)}\n')
                    i += 1

    def destroy(self) -> None:
            """Deallocates all loaded models and clears all data from the Calculator object."""
            self.loadedModels.clear()
            self.toRun.clear()
            self.descriptorData.clear()
            self.predictions.clear()
            self.inputFiles.clear()
            gc.collect()
            print("Calculator and all loaded models deallocated. All data cleared.")
            del self


# ************************  SATELLITE FUNCTIONS  ************************
            
def ward2ks2022(ward2017: np.ndarray) -> np.ndarray:
    """Converts a ``Ward2017`` descriptor to a ``KS2022`` descriptor (which is its subset).

    Args:
        ward2017: ``Ward2017`` descriptor. Must be a 1D ``np.ndarray`` of length ``271``.

    Returns:
        ``KS2022`` descriptor array.
    """

    assert isinstance(ward2017, np.ndarray)
    assert ward2017.shape == (271,)
    ward2017split = np.split(ward2017, [12, 15, 121, 126, 258, 264, 268, 269, 271])
    ks2022 = np.concatenate((
        ward2017split[0],
        ward2017split[2],
        ward2017split[4],
        ward2017split[6],
        ward2017split[8]
    ), axis=-1, dtype=np.float32)

    return ks2022

def overwritePrototypeLibrary(prototypeLibrary: dict) -> None:
    """Destructively overwrites the prototype library with a custom one. Used by the ``appendPrototypeLibrary()`` function
    to persist its changes. The other main use it to restore the default one to the original state based on a backup
    made earlier (see tests for an example)."""
    yaml_customDumper = YAML()
    yaml_customDumper.top_level_colon_align = True

    with resources.files('pysipfenn.misc').joinpath('prototypeLibrary.yaml').open('w+') as f:
        # Restructure the prototype library back to the original format of a list of dictionaries
        print(prototypeLibrary)
        prototypeList = [
            {
                'name': key,
                'origin': value['origin'],
                'POSCAR': LiteralScalarString(str(value['POSCAR']))
            }
            for key, value in prototypeLibrary.items()]
        print(prototypeList)
        # Persist the prototype library
        yaml_customDumper.dump(prototypeList, f)
        print(f'Updated prototype library persisted to {f.name}')

# *** HELPERS ***
def string2prototype(c: Calculator, prototype: str) -> Structure:
    """Converts a prototype string to a pymatgen ``Structure`` object.

    Args:
        c: ``Calculator`` object with the ``prototypeLibrary``.
        prototype: Prototype string.

    Returns:
        ``Structure`` object.
    """
    assert isinstance(prototype, str), 'Prototype string must be a string.'
    assert prototype in c.prototypeLibrary, \
        f'Unrecognized magic string for baseStruct: {prototype}. Please use one of the recognized magic ' \
        f'strings: {list(c.prototypeLibrary.keys())} or provide a Structure object.'
    s: Structure = c.prototypeLibrary[prototype]['structure']
    assert s.is_valid(), f'Invalid structure: {s}'
    return s

# *** WRAPPERS ***
def wrapper_KS2022_dilute_generate_descriptor(args):
    """Wraps the ``KS2022_dilute.generate_descriptor`` function for parallel processing."""
    return KS2022_dilute.generate_descriptor(*args)

def wrapper_KS2022_randomSolutions_generate_descriptor(args):
    """Wraps the ``KS2022_randomSolutions.generate_descriptor`` function for parallel processing."""
    return KS2022_randomSolutions.generate_descriptor(*args)


# This file is part of pySIPFENN and is licensed under the terms of the LGPLv3 or later.
# Copyright (C) 2023 Adam M. Krajewski, Jonathan Siegel

"""This ``KS2022`` feature vector calculator is a **special-input converter** modification of our the base ``KS2022``. Unlike the
``KS2022_dilute`` which reduces the computation, this one is designed to take an anonymous ``Structure`` and ``Composition`` pair and 
obtain vales of the ``KS2022`` features at infinite random supercell size.

It does that by expanding the ensamble of local chemical environments by iteratively adding supercells of the structure until the
features and composition converge. If you use this code, plese cite (as in ``KS2022.cite()``):

- Adam M. Krajewski, Jonathan W. Siegel, Jinchao Xu, Zi-Kui Liu, "Extensible Structure-Informed Prediction of Formation Energy with 
  improved accuracy and usability employing neural networks", Computational Materials Science, Volume 208, 2022, 111254

The core purpose of this module is to calculate numpy ``ndarray`` with ``256`` features constructed by considering all local chemical 
environments existing in an atomic structure. Their list is available in the ``labels_KS2022.csv`` and will be discussed in our upcoming
publication (Spring 2024).
"""

# pySIPFENN (for handling prototype library with high-level API at lower-level here)
import pysipfenn

# Standard Library Imports
import random
import math
import time
from collections import Counter
from typing import List, Union, Tuple
from importlib import resources

# Third Party Dependencies
from tqdm.contrib.concurrent import process_map
import numpy as np
from pymatgen.core import Structure, Element, Composition, PeriodicSite
from pymatgen.analysis.local_env import VoronoiNN

# Certain hard-coded basic elemental properties used in the featurization (attribute_matrix is compatible with Magpie references,
# and maxFeaturesInOQMD is based on the 2017 snapshot of OQMD, which was current when we started and will be retained in KS2022, but
# change in the future).
periodic_table_size = 112
f = resources.files('pysipfenn.descriptorDefinitions').joinpath("element_properties_Ward2017KS2022.csv")
attribute_matrix = np.loadtxt(f, delimiter=',')
attribute_matrix = np.nan_to_num(attribute_matrix)
attribute_matrix = attribute_matrix[:,[45, 33, 2, 32, 5, 48, 6, 10, 44, 42, 38, 40, 36, 43, 41, 37, 39, 35, 18, 13, 17]]
maxFeaturesInOQMD = np.array([
    13.1239, 5.01819, 12.0, 35.9918, 0.305284, 1.0, 1.89776, 0.61604, 0.251582, 0.505835, 0.671142, 0.648262, 0.74048, 
    89.0, 26.4054, 89.0, 93.0, 90.2828, 95.0, 30.0847, 95.0, 95.0, 90.6205, 233.189, 68.4697, 233.189, 242.992, 235.263, 
    3631.95, 1395.25, 3631.95, 3808.99, 3785.18, 16.0, 5.77718, 16.0, 16.0, 16.0, 5.99048, 1.66782, 5.99048, 6.0, 5.8207,
    213.0, 56.9856, 213.0, 213.0, 194.519, 3.19, 0.971532, 3.19, 3.19, 3.01435, 2.0, 0.749908, 2.0, 2.0, 2.0, 5.0, 1.80537, 
    5.0, 5.25839, 5.0, 10.0, 3.85266, 10.0, 10.0, 10.0, 14.0, 5.61254, 14.0, 14.0, 14.0, 26.0, 8.0159, 26.0, 28.0, 27.0, 
    1.0, 0.406792, 1.0, 1.0, 1.0, 5.0, 1.56393, 5.0, 5.0, 5.0, 9.0, 3.49493, 9.0, 9.0, 9.0, 13.0, 4.6277, 13.0, 13.0, 13.0,
    20.0, 6.30236, 20.0, 22.0, 22.0, 109.15, 36.0677, 109.15, 110.125, 108.956, 7.853, 2.78409, 7.853, 7.853, 7.853, 2.09127, 
    0.82173, 2.09127, 2.11066, 2.11066, 7.0, 1.0, 1.0, 1.0, 1.0, 1.0, 94.0, 93.0, 46.5, 94.0, 94.0, 94.0, 102.0, 97.0, 47.5, 
    102.0, 102.0, 102.0, 244.0, 242.992, 121.496, 244.0, 244.0, 244.0, 3823.0, 3808.99, 1904.5, 3823.0, 3823.0, 3823.0, 18.0, 
    17.0, 8.0, 18.0, 18.0, 18.0, 7.0, 6.0, 3.0, 7.0, 7.0, 7.0, 244.0, 213.0, 106.5, 244.0, 244.0, 244.0, 3.98, 3.19, 1.595, 
    3.98, 3.98, 3.98, 2.0, 2.0, 1.0, 2.0, 2.0, 2.0, 6.0, 6.0, 2.5, 6.0, 6.0, 6.0, 10.0, 10.0, 5.0, 10.0, 10.0, 10.0, 14.0, 
    14.0, 7.0, 14.0, 14.0, 14.0, 29.0, 28.0, 14.0, 29.0, 29.0, 29.0, 1.0, 1.0, 0.5, 1.0, 1.0, 1.0, 5.0, 5.0, 2.5, 5.0, 5.0, 
    5.0, 9.0, 9.0, 4.5, 9.0, 9.0, 9.0, 13.0, 13.0, 6.5, 13.0, 13.0, 13.0, 22.0, 22.0, 11.0, 22.0, 22.0, 22.0, 115.765, 110.125, 
    55.0625, 115.765, 115.765, 115.765, 7.853, 7.853, 3.9265, 7.853, 7.853, 7.853, 2.11066, 2.11066, 1.05533, 2.11066, 2.11066,
    2.11066, 1.0, 0.714286, 1.0, 0.875, 0.92145, 0.460725])


def local_env_function(
    local_env: dict,
    site: PeriodicSite
) -> List[np.ndarray]:
    """A prototype function which computes a weighted average over neighbors, weighted by the area of the Voronoi cell
    between them. This allows concurrently capturing impact of neighbor-neighbor interactions and geometric effects. 
    Critically, in contrast to cut-off based methods, the interaction is `guaranteed` to be continous as a function of 
    displacement.

    Args:
        local_env: A dictionary of the local environment of a site, as returned by a ``VoronoiNN`` generator. Contains 
            a number of critical geometric attributes like face distances, face areas, and corresponding face-bound volumes.
        site: The ``Site`` number for which the local environment is being computed.

    Returns:
        A nested list of ``np.ndarray``s. Contains several geometric attributes concatenated with gometry weighted neighbor-neighbor
        elemental attributes, and (2) a list of ``np.ndarray`` of geometry independent elemental attributes of the site.
    """
    local_attributes = np.zeros(attribute_matrix.shape[1])
    for key, value in site.species.get_el_amt_dict().items():
        local_attributes += value * attribute_matrix[Element(key).Z - 1, :]
    diff_attributes = np.zeros(attribute_matrix.shape[1])
    total_weight = 0
    volume = 0
    for _, neighbor_site in local_env.items():
        neighbor_attributes = np.zeros(attribute_matrix.shape[1])
        for key, value in neighbor_site['site'].species.get_el_amt_dict().items():
            neighbor_attributes += value * attribute_matrix[Element(key).Z - 1, :]
        diff_attributes += np.abs(local_attributes - neighbor_attributes) * neighbor_site['area']
        total_weight += neighbor_site['area']
        volume += neighbor_site['volume']
    elemental_properties_attributes = [diff_attributes / total_weight, local_attributes]
    # Calculate coordination number attribute
    average = 0
    variance = 0
    for neighbor_site in local_env.values():
        average += neighbor_site['area']
        variance += neighbor_site['area'] * neighbor_site['area']
    eff_coord_num = average * average / variance
    # Calculate Bond Length Attributes
    # AVG
    blen_average = 0
    for neighbor_site in local_env.values():
        blen_average += neighbor_site['area'] * 2 * neighbor_site['face_dist']
    blen_average /= total_weight
    # VAR
    blen_var = 0
    for neighbor_site in local_env.values():
        blen_var += neighbor_site['area'] * abs(2 * neighbor_site['face_dist'] - blen_average)
    blen_var /= total_weight * blen_average
    # Calculate Packing Efficiency info
    sphere_rad = min(neighbor_site['face_dist'] for neighbor_site in local_env.values())
    sphere_volume = (4.0 / 3.0) * math.pi * math.pow(sphere_rad, 3.0)
    return [np.concatenate(
        ([eff_coord_num, blen_average, blen_var, volume, sphere_volume], elemental_properties_attributes[0])),
        elemental_properties_attributes[1]]


class LocalAttributeGenerator:
    """A wrapper class which contains an instance of an NN generator (the default is a ``VoronoiNN``), a structure, and
    a function which computes the local environment attributes.
    
    Args:
        struct: A pymatgen ``Structure`` object.
        local_env_func: A function which computes the local environment attributes for a given site.
        nn_generator: A ``VoronoiNN`` generator object.
    """

    def __init__(
        self, 
        struct: Structure,
        local_env_func,
        nn_generator: VoronoiNN = VoronoiNN(
            compute_adj_neighbors=False, 
            extra_nn_info=False)
        ):
        self.generator = nn_generator
        self.struct = struct
        self.function = local_env_func

    def generate_local_attributes(self, n: int):
        """Wrapper pointing to a given ``Site`` index.
        
        Args:
            n: The index of the site for which the local environment attributes are being computed.
            
        Returns:
            A list of the local environment attributes for the site. The type will depend on the function used to compute the
            attributes. By default, this is a list of two numpy arrays computed by ``local_env_function``.
        """
        local_env = self.generator.get_voronoi_polyhedra(self.struct, n)
        return self.function(local_env, self.struct[n])


def generate_voronoi_attributes(
    struct: Structure, 
    local_funct=local_env_function
    ) -> tuple[np.ndarray, np.ndarray]:
    """Generates the local environment attributes for a given structure using a VoronoiNN generator.

    Args:
        struct: A pymatgen ``Structure`` object.
        local_funct: A function which computes the local environment attributes for a given site. By default, this is
            the prototype function ``local_env_function``, but you can neatly customize this to your own needs at this 
            level, if you so desire (e.g. to use a compiled alternative you have written).
            
    Returns:
        A tuple of two numpy arrays. Each contains concatenated outputs of respecive tuples from ``local_env_function``. Please note
        that, at this stage, the order of rows `does not` have to correspond to the order of sites in the structure and usually does not.
    """
    local_generator = LocalAttributeGenerator(struct, local_funct)
    attribute_list = list(
        map(local_generator.generate_local_attributes, 
            range(len(struct.sites))))
    return np.array([value[0] for value in attribute_list]), np.array([value[1] for value in attribute_list])


def most_common(
    attribute_properties: np.ndarray
    ) -> np.ndarray:
    """Calculates the attributes corresponding to the most common elements.
    
    Args:
        attribute_properties: A numpy array of the local environment attributes generated from ``generate_voronoi_attributes``.
        
    Returns:
        A numpy array of the attributes corresponding to the most common elements.
    """
    scores = np.unique(np.ravel(attribute_properties[:, 0]))  # get all unique atomic numbers
    max_occurrence = 0
    top_elements = []
    for score in scores:
        template = (attribute_properties[:, 0] == score)
        count = np.expand_dims(np.sum(template, 0), 0)[0]
        if count > max_occurrence:
            top_elements.clear()
            top_elements.append(score)
            max_occurrence = count
        elif count == max_occurrence:
            top_elements.append(score)
    output = np.zeros_like(attribute_properties[0, :])
    for elem in top_elements:
        output += attribute_matrix[int(elem) - 1, :]
    return output / len(top_elements)


def generate_descriptor(struct: Structure,
                        comp: Composition,
                        minimumSitesPerExpansion: int = 50,
                        featureConvergenceCriterion: float = 0.005,
                        compositionConvergenceCriterion: float = 0.01,
                        minimumElementOccurrences: int = 10,
                        plotParameters: bool = False,
                        printProgress: bool = True,
                        returnMeta: bool = False,
                        ) -> Union[np.ndarray, Tuple[np.ndarray, dict]]:
    """**Main functionality.** Generates the KS2022 descriptor for a given composition randomly distributed on a given
    structure until the convergence criteria are met. The descriptor is **KS2022** which is compatible with all KS2022
    models. It approaches values that would be reached by infinite supercell size.
    
    Args:
        struct: A pymatgen `Structure` object that will be used as the basis for the structure to be generated. It can
            be occupied by any species without affecting the result since all will be replaced by the composition.
        comp: A pymatgen `Composition` object that will be randomly distributed on the structure within accuracy
            determined by the `compositionConvergenceCriterion`.
        minimumSitesPerExpansion: The minimum number of sites that the base structure will be expanded to (doubling
            dimension-by-dimension) before it is used as an expansion step in each iteration adding local chemical
            environment information to the global pool.
            Optimal value will depend on the number of species and their relative fractions in the composition.
            Generally, low values will result in slower convergence (<20ish) and too high values (>150ish) will result
            in slower computation. The default value is 50.
        featureConvergenceCriterion: **The maximum difference between any feature belonging to the current iteration
            (statistics based on the global ensemble of local chemical environments) and the previous two iterations
            (before the last expansion, and the one before that)** expressed as a fraction of the maximum value of each
            structure-dependent KS2022 feature found in the OQMD database at the time of SIPFENN creation
            (see `maxFeaturesInOQMD` array). The default value is 0.005, corresponding to 0.5% of the maximum value.
        compositionConvergenceCriterion: The maximum average difference between any element fraction belonging in the
            current composition (superposition of all expansions) and the target composition (comp). The default value
            is 0.01, corresponding to deviation depending on the number of elements in the composition.
        minimumElementOccurrences: The minimum number of times all elements must occur in the composition before it is
            considered converged. This is to prevent the algorithm from converging before very dilute elements have
            had a chance to occur. The default value is 10.
        plotParameters: If True, the convergence history will be plotted using plotly and, by default, will display as an 
            interactive plot in your default web browser, allowing you to zoom and pan. The figure below shows an example
            of such plot for a complex BCC 6-component high entropy alloy. The default value is False.
            
            .. image:: https://raw.githubusercontent.com/PhasesResearchLab/pySIPFENN/main/pysipfenn/descriptorDefinitions/assets/KS2022_randomSolution_ConvergencePlot.png
                :alt: KS2022_randomSolution_ConvergencePlot
                :width: 800
            
        printProgress: If True, the progress will be printed to the console. The default value is True.
        returnMeta: If True, a dictionary containing the convergence history will be returned in addition to the
            descriptor. The default value is False.

    Returns: By default, a numpy array containing the KS2022 descriptor. Please note the stochastic nature of the
    algorithm, and that the result may vary slightly between runs and parameters. If returnMeta is True,
    a tuple containing the descriptor and a dictionary containing the convergence history will be returned.
    """

    # Obtain the elemental frequencies
    elementalFrequencies = dict(comp.fractional_composition.get_el_amt_dict())

    # If the number of sites is lower than 50, keep doubling up the structure, iterating over one of the 3 dimensions
    # at a time until there are at least 50 sites.
    adjustedStruct = struct.copy()
    i = 0
    while len(adjustedStruct.sites) < minimumSitesPerExpansion:
        scaling = [1, 1, 1]
        scaling[i] = 2
        adjustedStruct.make_supercell(scaling)
        i = (i + 1) % 3

    diff_properties = np.ndarray(shape=(0, 26))
    attribute_properties = np.ndarray(shape=(0, 21))
    propHistory = []
    diffHistory = []
    allOccupations = []
    maxDiff = 5
    compositionDistance = 1
    minOccupationCount = 0
    properties: np.ndarray = None
    currentComposition: Composition = None

    if printProgress:
        print(f'#Atoms | Comp. Distance AVG | Convergence Crit. MAX | Occupation Count MIN')

    if maxDiff < featureConvergenceCriterion:
        raise AssertionError('Invalid convergence criteria (maxDiff < featureConvergenceCriterion).')
    if compositionDistance < compositionConvergenceCriterion:
        raise AssertionError('Invalid convergence criteria (compositionDistance > compositionConvergenceCriterion).')
    if minOccupationCount > minimumElementOccurrences:
        raise AssertionError('Invalid convergence criteria (minOccupationCount > minimumElementOccurrences).')

    while maxDiff > featureConvergenceCriterion \
            or compositionDistance > compositionConvergenceCriterion \
            or minOccupationCount < minimumElementOccurrences:
        # Choose random structure occupation
        randomOccupation = random.choices(list(elementalFrequencies.keys()),
                                          weights=elementalFrequencies.values(),
                                          k=adjustedStruct.num_sites)
        allOccupations += randomOccupation
        occupationCount = dict(Counter(allOccupations))
        minOccupationCount = min(occupationCount.values())
        currentComposition = Composition.from_dict(occupationCount)
        # Adjust current elemental frequencies to push the current composition towards the target composition

        compositionDistance = 0
        for element in elementalFrequencies.keys():
            difference = currentComposition.fractional_composition.get_atomic_fraction(element) \
                         - comp.fractional_composition.get_atomic_fraction(element)
            compositionDistance += abs(difference)
            elementalFrequencies[element] -= difference * 0.1
        compositionDistance /= len(elementalFrequencies.keys())

        for site, occupation in zip(adjustedStruct.sites, randomOccupation):
            site.species = occupation

        diff_properties_instance, attribute_properties_instance = generate_voronoi_attributes(struct=adjustedStruct)

        diff_properties = np.concatenate((diff_properties, diff_properties_instance), axis=0)
        attribute_properties = np.concatenate((attribute_properties, attribute_properties_instance), axis=0)

        properties = np.concatenate(
            (np.stack(
                (np.mean(diff_properties, axis=0),
                 np.mean(np.abs(diff_properties - np.mean(diff_properties, axis=0)), axis=0),
                 np.min(diff_properties, axis=0),
                 np.max(diff_properties, axis=0),
                 np.max(diff_properties, axis=0) - np.min(diff_properties, axis=0)), axis=-1).reshape((-1)),
             np.stack(
                 (np.mean(attribute_properties, axis=0),
                  np.max(attribute_properties, axis=0) - np.min(attribute_properties, axis=0),
                  np.mean(np.abs(attribute_properties - np.mean(attribute_properties, axis=0)), axis=0),
                  np.max(attribute_properties, axis=0),
                  np.min(attribute_properties, axis=0),
                  most_common(attribute_properties)), axis=-1).reshape((-1))))
        # Normalize Bond Length properties.
        properties[6] /= properties[5]
        properties[7] /= properties[5]
        properties[8] /= properties[5]
        # Normalize the Cell Volume Deviation.
        properties[16] /= properties[15]
        # Remove properties not in magpie.
        properties = np.delete(properties, [4, 5, 9, 14, 15, 17, 18, 19, 21, 22, 23, 24])
        # Renormalize the packing efficiency.
        properties[12] *= adjustedStruct.num_sites / adjustedStruct.volume
        # Calculate and insert stoichiometry attributes.
        element_dict = currentComposition.fractional_composition.as_dict()
        position = 118
        for p in [10, 7, 5, 3, 2]:
            properties = np.insert(properties, position,
                                   math.pow(sum(math.pow(value, p) for value in element_dict.values()), 1.0 / p))
        properties = np.insert(properties, position, len(element_dict))
        # Calculate Valence Electron Statistics
        electron_occupation_dict = {'s': 0, 'p': 0, 'd': 0, 'f': 0}
        for key, value in element_dict.items():
            electron_occupation_dict['s'] += value * attribute_matrix[Element(key).Z - 1][8]
            electron_occupation_dict['p'] += value * attribute_matrix[Element(key).Z - 1][9]
            electron_occupation_dict['d'] += value * attribute_matrix[Element(key).Z - 1][10]
            electron_occupation_dict['f'] += value * attribute_matrix[Element(key).Z - 1][11]
        total_valence_factor = sum([val for (key, val) in electron_occupation_dict.items()])
        for orb in ['s', 'p', 'd', 'f']:
            properties = np.append(properties, electron_occupation_dict[orb] / total_valence_factor)
        # Calculate ionic compound attributes.
        max_ionic_char = 0
        av_ionic_char = 0
        for key1, value1 in element_dict.items():
            for key2, value2 in element_dict.items():
                ionic_char = 1.0 - math.exp(-0.25 * (Element(key1).X - Element(key2).X) ** 2)
                if ionic_char > max_ionic_char:
                    max_ionic_char = ionic_char
                av_ionic_char += ionic_char * value1 * value2
        properties = np.append(properties, max_ionic_char)
        properties = np.append(properties, av_ionic_char)
        properties = properties.astype(np.float32)

        propHistory.append(properties)
        # Calculate the difference between the current step and the previous step and divide it by maximum value of
        # each feature found in OQMD to normalize the difference.
        if len(propHistory) > 2:
            # Current iteration diff
            diff = np.subtract(properties, propHistory[-2])
            diff /= maxFeaturesInOQMD
            diffHistory.append(diff)
            # Calculate the additional diff to one level older iteration
            diff2 = np.subtract(properties, propHistory[-3])
            diff2 /= maxFeaturesInOQMD
            # Calculate the maximum difference across both differences
            maxDiff = max(np.concatenate((diff, diff2), axis=0))
            if printProgress:
                print(f'{attribute_properties.shape[0]:^6} | '
                      f'{compositionDistance: 18.6f} | '
                      f'{maxDiff: 21.6f} | '
                      f'{minOccupationCount:^4}')
        else:
            if printProgress:
                print(f'{attribute_properties.shape[0]:^6} | '
                      f'{compositionDistance: 18.6f} | '
                      f'{"(init)":^21} | '
                      f'{minOccupationCount:^4}')
    # ^^^ End of the long while-loop above

    if plotParameters:
        import plotly.express as px
        import pandas as pd
        import warnings
        warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)

        diffArray = np.array(diffHistory)

        # Plot the parameters as lines. Add hover text to show the parameter name based on the labels_KS2022.csv file.
        with resources.files('pysipfenn.descriptorDefinitions').joinpath('labels_KS2022.csv').open() as f:
            labels = f.readlines()
        fig = px.line(pd.DataFrame(diffArray, columns=labels), title='KS2022 Descriptor Parameters',
                      range_y=[-0.5, 0.5])
        # Add horizontal lines at convergence criteria of +/- 0.01.
        fig.add_hline(y=featureConvergenceCriterion, line_dash='dash', line_color='red')
        fig.add_hline(y=-featureConvergenceCriterion, line_dash='dash', line_color='red')

        fig.show()

    # print(f'Target: {comp.fractional_composition}')
    # print(f'Final:  {currentComposition.fractional_composition}')
    if properties is not None:
        assert properties.shape == (256,)
        assert isinstance(properties, np.ndarray)
        if returnMeta:
            return properties, {
                'diffHistory': diffHistory,
                'propHistory': propHistory,
                'finalAtomsN': attribute_properties.shape[0],
                'finalCompositionDistance': compositionDistance,
                'finalComposition': currentComposition.fractional_composition
            }
        else:
            return properties
    else:
        raise RuntimeError('KS2022_randomSolution descriptor failed to converge.')


def cite() -> List[str]:
    """Citation/s for the descriptor."""
    return [
        'Adam M. Krajewski, Jonathan W. Siegel, Jinchao Xu, Zi-Kui Liu, Extensible Structure-Informed Prediction of '
        'Formation Energy with improved accuracy and usability employing neural networks, Computational '
        'Materials Science, Volume 208, 2022, 111254'
    ]


def onlyStructural(descriptor: np.ndarray) -> np.ndarray:
    """Returns only the **part of the KS2022 descriptor that has to depend on structure**, useful in cases where the descriptor is used 
    as a fingerprint to compare polymorphs of the same compound. **Please note, this does not mean it selects all structure-dependent 
    features which span nearly entire descriptor, but only the part of the descriptor which is explicitly structure-dependent.** 

    Args:
        descriptor: A ``256``-length numpy ``ndarray`` of the KS2022 descriptor. Generated by the ``generate_descriptor`` function.

    Returns:
        A ``103``-length numpy ``ndarray`` of the structure-dependent part of the KS2022 descriptor. 
    """
    assert isinstance(descriptor, np.ndarray)
    assert descriptor.shape == (256,)
    descriptorSplit = np.split(descriptor, [68, 73, 93, 98, 113])
    ks2022_structural = np.concatenate((
        descriptorSplit[0],
        descriptorSplit[2],
        descriptorSplit[4]
    ), axis=-1, dtype=np.float32)
    assert ks2022_structural.shape == (103,)

    return ks2022_structural


def profile(test: str = 'FCC',
            comp: Composition = Composition('Cr12.8 Fe12.8 Co12.8 Ni12.8 Cu12.8 Al35.9'),
            nIterations: int = 1,
            plotParameters: bool = False,
            returnDescriptorAndMeta: bool = False) -> Union[None, Tuple[np.ndarray, dict]]:
    """Profiles the descriptor in parallel using one of the test structures.

    Args:
        test: The test structure to use. Options are 'FCC', 'BCC', and 'HCP'.
        comp: The composition to use. Should be a Composition pymatgen object.
        nIterations: The number of iterations to run. If 1, the descriptor will be calculated once and may be available
            in the return value. If >1, the descriptor will be calculated nIterations times and the result will
            be persisted in `f'TestResult_KS2022_randomSolution_{test}_{nIterations}iter.csv'` file in the current
            working directory.
        plotParameters: If True, the convergence history will be plotted using plotly. The default value is False.
        returnDescriptorAndMeta: If True, a tuple containing the descriptor and a dictionary containing the convergence
            history will be returned. The default value is False.

    Returns:
        Depending on the value of nIterations and returnDescriptorAndMeta, the return value will be either a tuple of
        the descriptor and a dictionary containing the convergence history, or None. In either case, the descriptor
        will be persisted in `f'TestResult_KS2022_randomSolution_{test}_{nIterations}iter.csv'` file.
    """
    c = pysipfenn.Calculator(autoLoad=False)

    try:
        s = c.prototypeLibrary[test]['structure']
    except KeyError:
        raise NotImplementedError(f'Unrecognized test name: {test}')

    name = f'TestResult_KS2022_randomSolution_{test}_{nIterations}iter.csv'

    if nIterations == 1:
        d, meta = generate_descriptor(s, comp, plotParameters=plotParameters, returnMeta=True)
        print(f"Got meta with :{meta.keys()} keys")
        with open(name, 'w+') as f:
            f.writelines([f'{v}\n' for v in d])
        if returnDescriptorAndMeta:
            return d, meta
    elif nIterations > 1:
        print(f'Running {nIterations} iterations in parallel...')
        d = process_map(generate_descriptor,
                        [s for _ in range(nIterations)],
                        [comp for _ in range(nIterations)],
                        chunksize=1,
                        max_workers=8)
        with open(name, 'w+') as f:
            f.writelines([f'{",".join([str(v) for v in di])}\n' for di in d])
        return None
    else:
        print('No descriptors generated.')
        return None

    print('Done!')


if __name__ == "__main__":
    print('You are running the KS2022_randomSolutions.py file directly. It is intended to be used as a module. '
          'A profiling task will now commence, going over several cases. This will take a while.')
    t0 = time.time()
    profile(test='FCC', plotParameters=True)
    profile(test='BCC', plotParameters=True)
    profile(test='HCP', plotParameters=True)
    profile(test='BCC', nIterations=6)
    print(f"All profiling tasks completed in {time.time() - t0:.2f} seconds. The results have been saved to the current working directory.")
    print(f"Average of {((time.time() - t0) / 4):.2f} seconds per task.")



# This file is part of pySIPFENN and is licensed under the terms of the LGPLv3 or later.
# Copyright (C) 2023 Adam M. Krajewski, Jonathan Siegel

"""This ``KS2022`` feature vector calculator is a highly optimized modification of our previous work calculating ``Ward2017`` features. It 
is **improved in terms of both feature selection and speed**. 

Most critically, it uses symmetry analysis to avoid redundant calculations of chemically equivalent atoms,
typically speeding up the calculation by a factor of 3 to 10 for ordered structures. Feature-optimization was also carried out to remove
several unphysical features or representations, such as space group being represented as a real number, rather than category. If
you use this code, plese cite (as in ``KS2022.cite()``):

- Adam M. Krajewski, Jonathan W. Siegel, Jinchao Xu, Zi-Kui Liu, "Extensible Structure-Informed Prediction of Formation Energy with 
  improved accuracy and usability employing neural networks", Computational Materials Science, Volume 208, 2022, 111254

The core purpose of this module is to calculate numpy ``ndarray`` with ``256`` features constructed by considering all local chemical 
environments existing in an atomic structure. Their list is available in the ``labels_KS2022.csv`` and will be discussed in our upcoming
publication (Spring 2024).
"""

# Standard Library Imports
import math
import time
import json
from collections import Counter
from typing import List, Dict
from importlib import resources

# Third Party Dependencies
from tqdm import tqdm
import numpy as np
from pymatgen.core import Structure, Element, PeriodicSite
from pymatgen.analysis.local_env import VoronoiNN
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

# Certain hard-coded basic elemental properties used in the featurization (compatible with Magpie references).
periodic_table_size = 112
f = resources.files('pysipfenn.descriptorDefinitions').joinpath("element_properties_Ward2017KS2022.csv")
attribute_matrix = np.loadtxt(f, delimiter=',')
attribute_matrix = np.nan_to_num(attribute_matrix)
attribute_matrix = attribute_matrix[:,[45, 33, 2, 32, 5, 48, 6, 10, 44, 42, 38, 40, 36, 43, 41, 37, 39, 35, 18, 13, 17]]


def local_env_function(
    local_env: dict,
    site: PeriodicSite
) -> List[np.ndarray]:
    """A prototype function which computes a weighted average over neighbors, weighted by the area of the Voronoi cell
    between them. This allows concurrently capturing impact of neighbor-neighbor interactions and geometric effects. 
    Critically, in contrast to cut-off based methods, the interaction is `guaranteed` to be continous as a function of 
    displacement.

    Args:
        local_env: A dictionary of the local environment of a site, as returned by a ``VoronoiNN`` generator. Contains 
            a number of critical geometric attributes like face distances, face areas, and corresponding face-bound volumes.
        site: The ``Site`` number for which the local environment is being computed.

    Returns:
        A nested list of ``np.ndarray``s. Contains several geometric attributes concatenated with gometry weighted neighbor-neighbor
        elemental attributes, and (2) a list of ``np.ndarray`` of geometry independent elemental attributes of the site.
    """
    local_attributes = np.zeros(attribute_matrix.shape[1])
    for key, value in site.species.get_el_amt_dict().items():
        local_attributes += value * attribute_matrix[Element(key).Z - 1, :]
    diff_attributes = np.zeros(attribute_matrix.shape[1])
    total_weight = 0
    volume = 0
    for _, neighbor_site in local_env.items():
        neighbor_attributes = np.zeros(attribute_matrix.shape[1])
        for key, value in neighbor_site['site'].species.get_el_amt_dict().items():
            neighbor_attributes += value * attribute_matrix[Element(key).Z - 1, :]
        diff_attributes += np.abs(local_attributes - neighbor_attributes) * neighbor_site['area']
        total_weight += neighbor_site['area']
        volume += neighbor_site['volume']
    elemental_properties_attributes = [diff_attributes / total_weight, local_attributes]
    # Calculate coordination number attribute
    average = 0
    variance = 0
    for neighbor_site in local_env.values():
        average += neighbor_site['area']
        variance += neighbor_site['area'] * neighbor_site['area']
    eff_coord_num = average * average / variance
    # Calculate Bond Length Attributes
    # AVG
    blen_average = 0
    for neighbor_site in local_env.values():
        blen_average += neighbor_site['area'] * 2 * neighbor_site['face_dist']
    blen_average /= total_weight
    # VAR
    blen_var = 0
    for neighbor_site in local_env.values():
        blen_var += neighbor_site['area'] * abs(2 * neighbor_site['face_dist'] - blen_average)
    blen_var /= total_weight * blen_average
    # Calculate Packing Efficiency info
    sphere_rad = min(neighbor_site['face_dist'] for neighbor_site in local_env.values())
    sphere_volume = (4.0 / 3.0) * math.pi * math.pow(sphere_rad, 3.0)
    return [np.concatenate(
        ([eff_coord_num, blen_average, blen_var, volume, sphere_volume], elemental_properties_attributes[0])),
        elemental_properties_attributes[1]]


class LocalAttributeGenerator:
    """A wrapper class which contains an instance of an NN generator (the default is a ``VoronoiNN``), a structure, and
    a function which computes the local environment attributes.
    
    Args:
        struct: A pymatgen ``Structure`` object.
        local_env_func: A function which computes the local environment attributes for a given site.
        nn_generator: A ``VoronoiNN`` generator object.
    """

    def __init__(
        self, 
        struct: Structure,
        local_env_func,
        nn_generator: VoronoiNN = VoronoiNN(
            compute_adj_neighbors=False, 
            extra_nn_info=False)
        ):
        self.generator = nn_generator
        self.struct = struct
        self.function = local_env_func

    def generate_local_attributes(self, n: int):
        """Wrapper pointing to a given ``Site`` index.
        
        Args:
            n: The index of the site for which the local environment attributes are being computed.
            
        Returns:
            A list of the local environment attributes for the site. The type will depend on the function used to compute the
            attributes. By default, this is a list of two numpy arrays computed by ``local_env_function``.
        """
        local_env = self.generator.get_voronoi_polyhedra(self.struct, n)
        return self.function(local_env, self.struct[n])


def generate_voronoi_attributes(
    struct: Structure, 
    local_funct=local_env_function
    ) -> tuple[np.ndarray, np.ndarray]:
    """Generates the local environment attributes for a given structure using a VoronoiNN generator. **Critically, this 
    implementation uses ``get_equivalentSitesMultiplicities`` to skip the expensive ``generate_local_attributes`` call when
    it is not guaranteed to be needed based on the symmetry of the structure.**

    Args:
        struct: A pymatgen ``Structure`` object.
        local_funct: A function which computes the local environment attributes for a given site. By default, this is
            the prototype function ``local_env_function``, but you can neatly customize this to your own needs at this 
            level, if you so desire (e.g. to use a compiled alternative you have written).
            
    Returns:
        A tuple of two numpy arrays. Each contains concatenated outputs of respecive tuples from ``local_env_function``. Please note
        that, at this stage, the order of rows `does not` have to correspond to the order of sites in the structure and usually does not.
    """
    local_generator = LocalAttributeGenerator(struct, local_funct)
    attribute_list = list()
    equivalentSitesMultiplicities = get_equivalentSitesMultiplicities(struct)
    for siteN in equivalentSitesMultiplicities:
        localAttributes = [local_generator.generate_local_attributes(siteN)]
        attribute_list += localAttributes * equivalentSitesMultiplicities[siteN]
    return np.array([value[0] for value in attribute_list]), np.array([value[1] for value in attribute_list])


def most_common(
    attribute_properties: np.ndarray
    ) -> np.ndarray:
    """Calculates the attributes corresponding to the most common elements.
    
    Args:
        attribute_properties: A numpy array of the local environment attributes generated from ``generate_voronoi_attributes``.
        
    Returns:
        A numpy array of the attributes corresponding to the most common elements.
    """
    scores = np.unique(np.ravel(attribute_properties[:, 0]))  # get all unique atomic numbers
    max_occurrence = 0
    top_elements = []
    for score in scores:
        template = (attribute_properties[:, 0] == score)
        count = np.expand_dims(np.sum(template, 0), 0)[0]
        if count > max_occurrence:
            top_elements.clear()
            top_elements.append(score)
            max_occurrence = count
        elif count == max_occurrence:
            top_elements.append(score)
    output = np.zeros_like(attribute_properties[0, :])
    for elem in top_elements:
        output += attribute_matrix[int(elem) - 1, :]
    return output / len(top_elements)


def get_equivalentSitesMultiplicities(
    struct: Structure,
    angle_tolerance: float = 0.1, 
    symprec: float = 0.001
    ) -> Dict[int, int]:
    """Takes a ``pymatgen`` ``Structure`` object, obtains an ``spglib`` symmetry dataset through the ``SpacegroupAnalyzer`` wrapper of 
    ```pymatgen```, and returns a dictionary of the multiplicities of equivalent sites.```
    
    Args:
        struct: A ``pymatgen`` ``Structure`` object.
        angle_tolerance: The angle tolerance for the symmetry analysis. By default, this is ``0.1``.
        symprec: The symmetry precision for the symmetry analysis. By default, this is ``0.001`` `when called from here`.
        
    Returns:
        A dictionary of the multiplicities of equivalent sites. The keys are the indices of the sites in the structure, and the values 
        are the integer multiplicities.
    """
    spgA = SpacegroupAnalyzer(
        struct, 
        angle_tolerance=angle_tolerance, 
        symprec=symprec)
    return dict(
        Counter(
            list(spgA.get_symmetry_dataset()['equivalent_atoms'])))


def generate_descriptor(struct: Structure) -> np.ndarray:
    """Main functionality sharing API with every other featurizer in ``pySIPEFNN``. Generates the KS2022 descriptor for a given structure.
    As explained in the top-level documentation, this descriptor is very efficient for ordered structures, and is minimizes expensive 
    computation by considering site equivalences due to symmetry.

    Args:
        struct: A pymatgen ``Structure`` object. It can be any ordered (e.g., crystal) or disordered (e.g., glass) structure with collapsed
            (defined) occupancies.
    Returns:
        A ``256``-length numpy ``ndarray`` of the descriptor. See ``labels_KS2022.csv`` for the meaning of each element of the array.
    """
    diff_properties, attribute_properties = generate_voronoi_attributes(struct)
    properties = np.concatenate(
        (np.stack(
            (np.mean(diff_properties, axis=0),
             np.mean(np.abs(diff_properties - np.mean(diff_properties, axis=0)), axis=0),
             np.min(diff_properties, axis=0),
             np.max(diff_properties, axis=0),
             np.max(diff_properties, axis=0) - np.min(diff_properties, axis=0)), axis=-1).reshape((-1)),
         np.stack(
             (np.mean(attribute_properties, axis=0),
              np.max(attribute_properties, axis=0) - np.min(attribute_properties, axis=0),
              np.mean(np.abs(attribute_properties - np.mean(attribute_properties, axis=0)), axis=0),
              np.max(attribute_properties, axis=0),
              np.min(attribute_properties, axis=0),
              most_common(attribute_properties)), axis=-1).reshape((-1))))
    # Normalize Bond Length properties.
    properties[6] /= properties[5]
    properties[7] /= properties[5]
    properties[8] /= properties[5]
    # Normalize the Cell Volume Deviation.
    properties[16] /= properties[15]
    # Remove properties not in magpie.
    properties = np.delete(properties, [4, 5, 9, 14, 15, 17, 18, 19, 21, 22, 23, 24])
    # Renormalize the packing efficiency.
    properties[12] *= len(attribute_properties) / struct.volume
    # Calculate and insert stoichiometry attributes.
    element_dict = struct.composition.fractional_composition.as_dict()
    position = 118
    for p in [10, 7, 5, 3, 2]:
        properties = np.insert(properties, position,
                               math.pow(sum(math.pow(value, p) for value in element_dict.values()), 1.0 / p))
    properties = np.insert(properties, position, len(element_dict))
    # Calculate Valence Electron Statistics
    electron_occupation_dict = {'s': 0, 'p': 0, 'd': 0, 'f': 0}
    total_valence_factor = 0
    for key, value in element_dict.items():
        electron_occupation_dict['s'] += value * attribute_matrix[Element(key).Z - 1][8]
        electron_occupation_dict['p'] += value * attribute_matrix[Element(key).Z - 1][9]
        electron_occupation_dict['d'] += value * attribute_matrix[Element(key).Z - 1][10]
        electron_occupation_dict['f'] += value * attribute_matrix[Element(key).Z - 1][11]
    total_valence_factor = sum([val for (key, val) in electron_occupation_dict.items()])
    for orb in ['s', 'p', 'd', 'f']:
        properties = np.append(properties, electron_occupation_dict[orb] / total_valence_factor)
    # Calculate ionic compound attributes.
    max_ionic_char = 0
    av_ionic_char = 0
    for key1, value1 in element_dict.items():
        for key2, value2 in element_dict.items():
            ionic_char = 1.0 - math.exp(-0.25 * (Element(key1).X - Element(key2).X) ** 2)
            if ionic_char > max_ionic_char:
                max_ionic_char = ionic_char
            av_ionic_char += ionic_char * value1 * value2
    properties = np.append(properties, max_ionic_char)
    properties = np.append(properties, av_ionic_char)
    properties = properties.astype(np.float32)
    return properties


def cite() -> List[str]:
    """Citation/s for the descriptor."""
    return [
        'Adam M. Krajewski, Jonathan W. Siegel, Jinchao Xu, Zi-Kui Liu, Extensible Structure-Informed Prediction of '
        'Formation Energy with improved accuracy and usability employing neural networks, Computational '
        'Materials Science, Volume 208, 2022, 111254'
        ]


def onlyStructural(descriptor: np.ndarray) -> np.ndarray:
    """Returns only the **part of the KS2022 descriptor that has to depend on structure**, useful in cases where the descriptor is used 
    as a fingerprint to compare polymorphs of the same compound. **Please note, this does not mean it selects all structure-dependent 
    features which span nearly entire descriptor, but only the part of the descriptor which is explicitly structure-dependent.** 

    Args:
        descriptor: A ``256``-length numpy ``ndarray`` of the KS2022 descriptor. Generated by the ``generate_descriptor`` function.

    Returns:
        A ``103``-length numpy ``ndarray`` of the structure-dependent part of the KS2022 descriptor. 
    """
    assert isinstance(descriptor, np.ndarray)
    assert descriptor.shape == (256,)
    descriptorSplit = np.split(descriptor, [68, 73, 93, 98, 113])
    ks2022_structural = np.concatenate((
        descriptorSplit[0],
        descriptorSplit[2],
        descriptorSplit[4]
    ), axis=-1, dtype=np.float32)
    assert ks2022_structural.shape == (103,)

    return ks2022_structural


def profile(
    test: str = 'JVASP-10001', 
    nRuns: int = 10,
    persistResult: bool = True
    ) -> None:
    """Profiles the descriptor in `series` using one of the test structures.
    
    Args:
        test: The name of the test structure. By default, this is ``'JVASP-10001'``. Currently implemented tests are: ``'JVASP-10001'`` and 
            ``'diluteNiAlloy'``.
        nRuns: The number of runs. By default, this is ``10``.
        persistResult: Whether to persist the result to a file (``'KS2022_TestResult.csv'``) to allow for inspection. By default, this is
            ``True``.
    """
    if test == 'JVASP-10001':
        print(
            f'KS2022 profiling/testing task will calculate a descriptor for Li2 Zr1 Te1 O6 (JVASP-10001) {nRuns} times in series.')
        matStr = '{"@module": "pymatgen.core.structure", "@class": "Structure", "charge": null, "lattice": {"matrix": [[4.599305652662459, 0.0098015076998823, 3.1052612865443736], [1.6553257726204653, 4.291108475854712, 3.1052602938979565], [0.0142541214919749, 0.0098025099996131, 5.549419141866351]], "a": 5.549446478152326, "b": 5.549446536179343, "c": 5.549446105810423, "alpha": 55.82714459985832, "beta": 55.82714014289371, "gamma": 55.82713972779092, "volume": 109.15484625642743}, "sites": [{"species": [{"element": "Li", "occu": 1.0}], "abc": [0.2738784872669924, 0.2738784872670407, 0.2738784872673032], "xyz": [1.7169128904007063, 1.1806114167777613, 3.220794775377278], "label": "Li", "properties": {}}, {"species": [{"element": "Li", "occu": 1.0}], "abc": [0.7852272010728069, 0.7852272010728856, 0.785227201073315], "xyz": [4.922499451739965, 3.3848887059434927, 9.234225338163633], "label": "Li", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.8669964454661124, 0.604089882092114, 0.241821769873143], "xyz": [4.990994160164061, 2.603083545876856, 5.910077181137658], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.717840894529788, 0.1213675889628683, 0.393537009186973], "xyz": [3.508082106234713, 0.531695063215412, 4.789863306469278], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.1213675889638402, 0.3935370091873943, 0.7178408945283384], "xyz": [1.2198707830817896, 1.6969362235910317, 5.58251292516964], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.3935370091861915, 0.7178408945293856, 0.1213675889634014], "xyz": [2.999987512595622, 3.085380109860344, 4.124637687962297], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.2418217698721573, 0.8669964454671221, 0.6040898820921513], "xyz": [2.555984564633329, 3.728667610729149, 6.795517372377343], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.6040898820933115, 0.2418217698723637, 0.8669964454664059], "xyz": [3.191046090145173, 1.0521031793025595, 7.4381031350436535], "label": "O", "properties": {}}, {"species": [{"element": "Te", "occu": 1.0}], "abc": [0.4965905610507353, 0.4965905610507355, 0.4965905610507361], "xyz": [3.113069390835793, 2.1406591357024984, 5.8398755612146624], "label": "Te", "properties": {}}, {"species": [{"element": "Zr", "occu": 1.0}], "abc": [0.0006501604980668, 0.0006501604980928, 0.0006501604982344], "xyz": [0.00407578174946036, 0.002802654981945192, 0.007645848918263076], "label": "Zr", "properties": {}}]}'
    elif test == 'diluteNiAlloy':
        print(
            f'KS2022 profiling/testing task will calculate a descriptor for a dilute FCC Ni31Cr1 alloy {nRuns} times in series.')
        matStr = '{"@module": "pymatgen.core.structure", "@class": "Structure", "charge": null, "lattice": {"matrix": [[6.995692, 0.0, 0.0], [0.0, 6.995692, 0.0], [0.0, 0.0, 6.995692]], "a": 6.995692, "b": 6.995692, "c": 6.995692, "alpha": 90.0, "beta": 90.0, "gamma": 90.0, "volume": 342.36711365619243}, "sites": [{"species": [{"element": "Cr", "occu": 1}], "abc": [0.0, 0.0, 0.0], "xyz": [0.0, 0.0, 0.0], "label": "Cr", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.0, 0.5], "xyz": [0.0, 0.0, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.5, 0.0], "xyz": [0.0, 3.497846, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.5, 0.5], "xyz": [0.0, 3.497846, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.0, 0.0], "xyz": [3.497846, 0.0, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.0, 0.5], "xyz": [3.497846, 0.0, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.5, 0.0], "xyz": [3.497846, 3.497846, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.5, 0.5], "xyz": [3.497846, 3.497846, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.25, 0.0], "xyz": [1.748923, 1.748923, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.25, 0.5], "xyz": [1.748923, 1.748923, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.7500000000000001, 0.0], "xyz": [1.748923, 5.2467690000000005, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.7500000000000001, 0.5], "xyz": [1.748923, 5.2467690000000005, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.25, 0.0], "xyz": [5.2467690000000005, 1.748923, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.25, 0.5], "xyz": [5.2467690000000005, 1.748923, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.7500000000000001, 0.0], "xyz": [5.2467690000000005, 5.2467690000000005, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.7500000000000001, 0.5], "xyz": [5.2467690000000005, 5.2467690000000005, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.0, 0.25], "xyz": [1.748923, 0.0, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.0, 0.7500000000000001], "xyz": [1.748923, 0.0, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.5, 0.25], "xyz": [1.748923, 3.497846, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.5, 0.7500000000000001], "xyz": [1.748923, 3.497846, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.0, 0.25], "xyz": [5.2467690000000005, 0.0, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.0, 0.7500000000000001], "xyz": [5.2467690000000005, 0.0, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.5, 0.25], "xyz": [5.2467690000000005, 3.497846, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.5, 0.7500000000000001], "xyz": [5.2467690000000005, 3.497846, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.25, 0.25], "xyz": [0.0, 1.748923, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.25, 0.7500000000000001], "xyz": [0.0, 1.748923, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.7500000000000001, 0.25], "xyz": [0.0, 5.2467690000000005, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.7500000000000001, 0.7500000000000001], "xyz": [0.0, 5.2467690000000005, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.25, 0.25], "xyz": [3.497846, 1.748923, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.25, 0.7500000000000001], "xyz": [3.497846, 1.748923, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.7500000000000001, 0.25], "xyz": [3.497846, 5.2467690000000005, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.7500000000000001, 0.7500000000000001], "xyz": [3.497846, 5.2467690000000005, 5.2467690000000005], "label": "Ni", "properties": {}}], "@version": null}'
    else:
        print('Unrecognized test name.')
        return None
    t0 = time.time()
    structList = [Structure.from_dict(json.loads(matStr))] * nRuns
    for s in tqdm(structList):
        d = generate_descriptor(s)
    if persistResult:
        with open('KS2022_TestResult.csv', 'w+') as f:
            f.writelines([f'{v}\n' for v in d])
    print(f"Done in {time.time() - t0} seconds.")
    print(f"Average time per run: {(time.time() - t0) / nRuns} seconds.")
    return None


def profileParallel(
    test: str = 'JVASP-10001', 
    nRuns: int = 1000,
    makeSupercell222: bool = False
    ) -> None:
    """Profiles the descriptor in `parallel` using one of the test structures.
    
    Args:
        test: The name of the test structure. By default, this is ``'JVASP-10001'``. Currently implemented tests are: ``'JVASP-10001'`` and 
            ``'diluteNiAlloy'``.
        nRuns: The number of total runs done in parallel by 8 workers. By default, this is ``1000``.
        makeSupercell222: Whether to make a 2x2x2 supercell of the structure before profiling, increasing the number of atoms by a factor
            of 8, but should not increase time thanks to the symmetry consierations. By default, this is ``False``.
    """
    from tqdm.contrib.concurrent import process_map
    if test == 'JVASP-10001':
        print(
            f'KS2022 profiling/testing task will calculate a descriptor for Li2 Zr1 Te1 O6 (JVASP-10001) {nRuns} times in parallel with 8 workers.')
        matStr = '{"@module": "pymatgen.core.structure", "@class": "Structure", "charge": null, "lattice": {"matrix": [[4.599305652662459, 0.0098015076998823, 3.1052612865443736], [1.6553257726204653, 4.291108475854712, 3.1052602938979565], [0.0142541214919749, 0.0098025099996131, 5.549419141866351]], "a": 5.549446478152326, "b": 5.549446536179343, "c": 5.549446105810423, "alpha": 55.82714459985832, "beta": 55.82714014289371, "gamma": 55.82713972779092, "volume": 109.15484625642743}, "sites": [{"species": [{"element": "Li", "occu": 1.0}], "abc": [0.2738784872669924, 0.2738784872670407, 0.2738784872673032], "xyz": [1.7169128904007063, 1.1806114167777613, 3.220794775377278], "label": "Li", "properties": {}}, {"species": [{"element": "Li", "occu": 1.0}], "abc": [0.7852272010728069, 0.7852272010728856, 0.785227201073315], "xyz": [4.922499451739965, 3.3848887059434927, 9.234225338163633], "label": "Li", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.8669964454661124, 0.604089882092114, 0.241821769873143], "xyz": [4.990994160164061, 2.603083545876856, 5.910077181137658], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.717840894529788, 0.1213675889628683, 0.393537009186973], "xyz": [3.508082106234713, 0.531695063215412, 4.789863306469278], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.1213675889638402, 0.3935370091873943, 0.7178408945283384], "xyz": [1.2198707830817896, 1.6969362235910317, 5.58251292516964], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.3935370091861915, 0.7178408945293856, 0.1213675889634014], "xyz": [2.999987512595622, 3.085380109860344, 4.124637687962297], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.2418217698721573, 0.8669964454671221, 0.6040898820921513], "xyz": [2.555984564633329, 3.728667610729149, 6.795517372377343], "label": "O", "properties": {}}, {"species": [{"element": "O", "occu": 1.0}], "abc": [0.6040898820933115, 0.2418217698723637, 0.8669964454664059], "xyz": [3.191046090145173, 1.0521031793025595, 7.4381031350436535], "label": "O", "properties": {}}, {"species": [{"element": "Te", "occu": 1.0}], "abc": [0.4965905610507353, 0.4965905610507355, 0.4965905610507361], "xyz": [3.113069390835793, 2.1406591357024984, 5.8398755612146624], "label": "Te", "properties": {}}, {"species": [{"element": "Zr", "occu": 1.0}], "abc": [0.0006501604980668, 0.0006501604980928, 0.0006501604982344], "xyz": [0.00407578174946036, 0.002802654981945192, 0.007645848918263076], "label": "Zr", "properties": {}}]}'
    elif test == 'diluteNiAlloy':
        print(
            f'KS2022 profiling/testing task will calculate a descriptor for a dilute FCC Ni31Cr1 alloy {nRuns} times in parallel with 8 workers.')
        matStr = '{"@module": "pymatgen.core.structure", "@class": "Structure", "charge": null, "lattice": {"matrix": [[6.995692, 0.0, 0.0], [0.0, 6.995692, 0.0], [0.0, 0.0, 6.995692]], "a": 6.995692, "b": 6.995692, "c": 6.995692, "alpha": 90.0, "beta": 90.0, "gamma": 90.0, "volume": 342.36711365619243}, "sites": [{"species": [{"element": "Cr", "occu": 1}], "abc": [0.0, 0.0, 0.0], "xyz": [0.0, 0.0, 0.0], "label": "Cr", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.0, 0.5], "xyz": [0.0, 0.0, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.5, 0.0], "xyz": [0.0, 3.497846, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.5, 0.5], "xyz": [0.0, 3.497846, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.0, 0.0], "xyz": [3.497846, 0.0, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.0, 0.5], "xyz": [3.497846, 0.0, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.5, 0.0], "xyz": [3.497846, 3.497846, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.5, 0.5], "xyz": [3.497846, 3.497846, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.25, 0.0], "xyz": [1.748923, 1.748923, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.25, 0.5], "xyz": [1.748923, 1.748923, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.7500000000000001, 0.0], "xyz": [1.748923, 5.2467690000000005, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.7500000000000001, 0.5], "xyz": [1.748923, 5.2467690000000005, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.25, 0.0], "xyz": [5.2467690000000005, 1.748923, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.25, 0.5], "xyz": [5.2467690000000005, 1.748923, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.7500000000000001, 0.0], "xyz": [5.2467690000000005, 5.2467690000000005, 0.0], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.7500000000000001, 0.5], "xyz": [5.2467690000000005, 5.2467690000000005, 3.497846], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.0, 0.25], "xyz": [1.748923, 0.0, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.0, 0.7500000000000001], "xyz": [1.748923, 0.0, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.5, 0.25], "xyz": [1.748923, 3.497846, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.25, 0.5, 0.7500000000000001], "xyz": [1.748923, 3.497846, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.0, 0.25], "xyz": [5.2467690000000005, 0.0, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.0, 0.7500000000000001], "xyz": [5.2467690000000005, 0.0, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.5, 0.25], "xyz": [5.2467690000000005, 3.497846, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.7500000000000001, 0.5, 0.7500000000000001], "xyz": [5.2467690000000005, 3.497846, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.25, 0.25], "xyz": [0.0, 1.748923, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.25, 0.7500000000000001], "xyz": [0.0, 1.748923, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.7500000000000001, 0.25], "xyz": [0.0, 5.2467690000000005, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.0, 0.7500000000000001, 0.7500000000000001], "xyz": [0.0, 5.2467690000000005, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.25, 0.25], "xyz": [3.497846, 1.748923, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.25, 0.7500000000000001], "xyz": [3.497846, 1.748923, 5.2467690000000005], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.7500000000000001, 0.25], "xyz": [3.497846, 5.2467690000000005, 1.748923], "label": "Ni", "properties": {}}, {"species": [{"element": "Ni", "occu": 1}], "abc": [0.5, 0.7500000000000001, 0.7500000000000001], "xyz": [3.497846, 5.2467690000000005, 5.2467690000000005], "label": "Ni", "properties": {}}], "@version": null}'
    else:
        print('Unrecognized test name.')
        return None
    t0 = time.time()
    s = Structure.from_dict(json.loads(matStr))
    if makeSupercell222:
        s.make_supercell(scaling_matrix=[2,2,2])
    structList = [s] * nRuns
    process_map(generate_descriptor, structList, max_workers=8)
    print(f"Done in {time.time() - t0} seconds.")
    print(f"Average time per run: {(time.time() - t0) / nRuns} seconds.")
    return None


if __name__ == "__main__":
    profile(test='JVASP-10001')
    profile(test='diluteNiAlloy')
    profileParallel(test='JVASP-10001')
    profileParallel(test='diluteNiAlloy')
